#include "CudaMoGEM.hpp"

#include <cassert>
#include <cstdio>
#include <vector>


#ifndef USE_CUBLAS_GEMM
#include "common/blas.cuh"
#endif /*USE_CUBLAS_GEMM*/

#include "common/common.cuh"
#include "common/reduce.cuh"
#include "common/statistics.cuh"
#include "config.hpp"
#include "ext/prob/prob.hpp"
#include "kernels.cuh"

#include <cublas_v2.h>

CudaMoGEM::CudaMoGEM(unsigned nfeature, unsigned nsample, unsigned nclass, cublasHandle_t ch):
	nfeature(nfeature), nsample(nsample), nclass(nclass), ch(ch) {
		if (ch != NULL) {
			CUDA_DIE_UNLESS_SUCCESS(cublasGetStream(ch, &cs));
		}
	}

void CudaMoGEM::initilize_GMDistribution(float *P, float *mu, float *Sigma, const float *samples, const float *ones, float *tmpVar) {
	assert(nclass == 1);

	float one = 1.0;
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(P, &one, sizeof(*P), cudaMemcpyHostToDevice));

#ifndef USE_CUBLAS_GEMM
	sum_rowwise(cs, mu, samples, dim3(nfeature,nsample), ones, tmpVar, 1.0/nsample);
#else
	sum_rowwise(ch, mu, samples, dim3(nfeature,nsample), ones, 1.0/nsample);
#endif /*USE_CUBLAS_GEMM*/
	cov_rowwise(ch, Sigma, tmpVar, samples, dim3(nfeature,nsample), mu);
}

void CudaMoGEM::print_gm(const float *P, const float *mu, const float *Sigma) const {
	float *hP = (float *)alloca(nclass*sizeof(*hP));
	float *hmu = (float *)alloca(nfeature*nclass*sizeof(*hmu));
	float *hSigma = (float *)alloca(nfeature*nfeature*nclass*sizeof(*hSigma));

	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hP, P, nclass*sizeof(*hP), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hmu, mu, nfeature*nclass*sizeof(*hmu), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hSigma, Sigma, nfeature*nfeature*nclass*sizeof(*hSigma), cudaMemcpyDeviceToHost));

	std::printf("%11s %11s %11s\n", "P", "mu", "Sigma");
	for (unsigned k=0; k < nclass; k++) {
		for (unsigned j=0; j<nfeature; j++) {
			if (j == 0) std::printf("%- 11.4g", hP[k]);
			else std::printf("%11s", "");

			std::printf(" %- 11.4g", hmu[k*nfeature + j]);
			for (unsigned i=0; i<nfeature; i++) {
				std::printf(" %- 11.4g", hSigma[k*nfeature*nfeature + j*nfeature + i]);
			}
			std::printf("\n");
		}
		std::printf("\n");
	}
}

void CudaMoGEM::pdfs_from_samples(float *pdfs, const float *samples, const float *P, const float *mu, float *luSigma, float *iSigma, int *pSigma, int *dbatchStatus, float **drefs) {
	float **hrefs = (float **)alloca(2*nclass*sizeof(*hrefs));
	for (unsigned i=0; i<nclass; i++) {
		hrefs[0*nclass + i] = luSigma + i*nfeature*nfeature;
		hrefs[1*nclass + i] = iSigma + i*nfeature*nfeature;
	}

	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(drefs, hrefs, 2*nclass*sizeof(*drefs), cudaMemcpyHostToDevice));
	float **drluSigma = drefs + 0*nclass;
	float **driSigma = drefs + 1*nclass;

	int *hbatchStatus = (int*)alloca(nclass*sizeof(hbatchStatus));
	CUDA_DIE_UNLESS_SUCCESS(cublasSgetrfBatched(ch, nfeature, drluSigma, nfeature, pSigma, dbatchStatus, nclass));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hbatchStatus, dbatchStatus, nclass*sizeof(*hbatchStatus), cudaMemcpyDeviceToHost));
	for (unsigned i=0; i<nclass; i++) { assert(hbatchStatus[i] == 0); }
	CUDA_DIE_UNLESS_SUCCESS(cublasSgetriBatched(ch, nfeature, (const float **)drluSigma, nfeature, pSigma, driSigma, nfeature, dbatchStatus, nclass));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hbatchStatus, dbatchStatus, nclass*sizeof(*hbatchStatus), cudaMemcpyDeviceToHost));
	for (unsigned i=0; i<nclass; i++) { assert(hbatchStatus[i] == 0); }

	dim3 p(nclass,nsample), ts(nclass, 256/nclass);
	compute_pdfs<<<gs(p,ts),ts,0,cs>>>(p, dim3(nfeature,nsample), pdfs, samples, P, mu, pSigma, luSigma, iSigma);
}

void CudaMoGEM::weights_from_pdfs(float *weights, const float *pdfs) {
	unsigned warp_width = 1 << (unsigned)ceil(log2((double)nclass));
	dim3 p(nclass,nsample), ts(warp_width, 256/warp_width);
	samplewise_normalization<<<gs(p,ts),ts,0,cs>>>(p, weights, pdfs);
}

void CudaMoGEM::classwise_share(float *P_with_respect_to_weights, const float *weights, const float *ones, float *P_to_be_summed, float scalar) {

#ifndef USE_CUBLAS_GEMM
	sum_rowwise(cs, P_with_respect_to_weights, weights, dim3(nclass,nsample), ones, P_to_be_summed, scalar);
#else
	sum_rowwise(ch, P_with_respect_to_weights, weights, dim3(nclass,nsample), ones, scalar);
#endif /*USE_CUBLAS_GEMM*/
}

void CudaMoGEM::mstep(float *P, float *mu, float *Sigma, const float *samples, const float *weights, float *FSC_A, float *FSC_B, float *FSC_C, const float *ones, const float *fallback_covariance) {
	dim3 p(nfeature,nsample), ts(nfeature, 256/nfeature), sweights(nclass,nsample);
#ifndef USE_CUBLAS_GEMM
	sum_rowwise(cs, P, weights, sweights, ones, FSC_A, 1.0/nsample);
#else
	sum_rowwise(ch, P, weights, sweights, ones, 1.0/nsample);
#endif /*USE_CUBLAS_GEMM*/

	{ // emphasize when temporary FSC_A is available
		float *weighted_samples = FSC_A;
		weight_vectorized_samples<<<gs(p,ts),ts,0,cs>>>(p, sweights, weighted_samples, samples, weights, P);
		sum_rowwise(cs, mu, weighted_samples, dim3(nfeature*nclass, nsample), ones, FSC_B);
	}

	float *mu_deviation_over_P = FSC_A;
	float *weighted_mu_deviation = FSC_B;
	populate_covariance_factors<<<gs(p,ts),ts,0,cs>>>(p, sweights, mu_deviation_over_P, weighted_mu_deviation, P, mu, samples, weights);

	std::vector<unsigned> singular_classes;
	float *hP = (float *)alloca(nclass*sizeof(*hP));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hP, P, nclass*sizeof(*hP), cudaMemcpyDeviceToHost));
	for (unsigned i=0; i<nclass; i++) {
		if (hP[i]*nsample >= 2) {
			float *A = mu_deviation_over_P + i*nsample*nfeature;
			float *B = weighted_mu_deviation + i*nsample*nfeature;
			float *C = Sigma + i*nfeature*nfeature;

			float alpha=1.0, beta=0.0;
			dim3 p(nfeature,nsample,nfeature), ts(32*32);
			unsigned bs = gs(dim3(nsample),ts).x;
#ifndef USE_CUBLAS_GEMM
			Sgemm1_nt<<<bs,ts,0,cs>>>(p, FSC_C, A, B); p.y = bs;
			Sgemm2<<<1,ts_warp_aligned(bs),0,cs>>>(p, beta, C, alpha, FSC_C);
#else
			CUDA_DIE_UNLESS_SUCCESS(cublasSgemm(ch, CUBLAS_OP_N, CUBLAS_OP_T, nfeature, nfeature, nsample, &alpha, A, nfeature, B, nfeature, &beta, C, nfeature));
#endif /*USE_CUBLAS_GEMM*/
		} else {
			singular_classes.push_back(i);
		}
	}

	for (auto const& i : singular_classes) {
		std::cout << i << "th class gone singular, randomizing" << std::endl;
		float newP = 1.0/nclass;
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(P + i, &newP, sizeof(*P), cudaMemcpyHostToDevice));
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(mu + i*nfeature, samples + (std::rand()%nsample)*nfeature, nfeature*sizeof(*mu), cudaMemcpyDeviceToDevice));
		dim3 p(nfeature*nfeature);
		scale_array<<<1,p,0,cs>>>(p, Sigma + i*nfeature*nfeature, fallback_covariance, 1.0F/nclass);
	}
}

void CudaMoGEM::score(float *scores, const float *mu, const float *Sigma, const float *samples, const float *weights, float *P_with_respect_to_weights, float *bin_bounds, float *weight_bins, float *P_to_be_summed, const float *ones) {
	float *hmu = (float *)alloca(nfeature*nclass*sizeof(*hmu));
	float *hSigma = (float *)alloca(nfeature*nfeature*nclass*sizeof(*hSigma));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hmu, mu, nfeature*nclass*sizeof(*hmu), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hSigma, Sigma, nfeature*nfeature*nclass*sizeof(*hSigma), cudaMemcpyDeviceToHost));

	classwise_share(P_with_respect_to_weights, weights, ones, P_to_be_summed, 1.0/nsample);

	float *hbin_bounds = (float *)alloca((NPROBABILITY_BINS+1)*nfeature*nclass*sizeof(*hbin_bounds));
	for (unsigned j=0; j<nclass; j++) {
		for (unsigned i=0; i<nfeature; i++) {
			for (unsigned bin=0; bin<=NPROBABILITY_BINS; bin++) {
				float bound = normal_cdf_inv((double)(bin) / NPROBABILITY_BINS, hmu[j*nfeature + i], sqrt(hSigma[j*nfeature*nfeature + i*nfeature+i]));
				hbin_bounds[j*nfeature*(NPROBABILITY_BINS+1) + i*(NPROBABILITY_BINS+1) + bin] = bound;
			}
		}
	}
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(bin_bounds, hbin_bounds, (NPROBABILITY_BINS+1)*nfeature*nclass*sizeof(*bin_bounds), cudaMemcpyHostToDevice));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemset(weight_bins, 0, NPROBABILITY_BINS*nfeature*nclass*sizeof(*weight_bins)));

	dim3 p(NPROBABILITY_BINS,nfeature,nsample/16), ts(NPROBABILITY_BINS, nfeature, 512/(nfeature*NPROBABILITY_BINS)), sweights(nclass,nsample);
	weight_binning<<<gs(p,ts),ts,0,cs>>>(p, sweights, weight_bins, bin_bounds, samples, weights);

	float cl = 1 - (1-CONFIDENCE_LEVEL)/(nfeature*nclass);
	float threshold = chi_square_cdf_inv(cl, NPROBABILITY_BINS-DOFS);

	assert(NPROBABILITY_BINS < 32);
	p = dim3(32,nfeature,nclass);
	scores_from_weight_bins<<<1,p,0,cs>>>(p, scores, weight_bins, P_with_respect_to_weights, threshold, nsample);
}

void CudaMoGEM::determine_least_fitting_class(unsigned *hout_iworst_score, float *hout_worst_score, const float *scores, unsigned *index, float *value) {
	dim3 p(nfeature*nclass), ts(32*32);
	ireduce_array<IRMAX><<<1,ts,0,cs>>>(p, index, value, scores);
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hout_iworst_score, index, sizeof(*hout_iworst_score), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hout_worst_score, value, sizeof(*hout_worst_score), cudaMemcpyDeviceToHost));
}

void CudaMoGEM::split_class(float *nP, float *nmu, float *nSigma, const float *P, const float *mu, const float *Sigma, unsigned iscore, float score) {
	unsigned iclass = iscore/nfeature, ifeature = iscore%nfeature;

	dim3 p(nfeature,nfeature,nclass+1);
	ksplit_class<<<1,p,0,cs>>>(p, nP, nmu, nSigma, P, mu, Sigma, iclass, ifeature);
	nclass++;
}

float CudaMoGEM::log_likelihood(const float *pdfs, float *partial_aggregates) {
	unsigned warp_width = 1 << (unsigned)ceil(log2((double)nclass));
	dim3 p(nclass,nsample), ts(warp_width, 256/warp_width);
	compute_log_likelihood<<<gs(p,ts),ts,0,cs>>>(p, partial_aggregates, pdfs);

	reduce_array<RSUM><<<gs(nsample,32*32),32*32,0,cs>>>(dim3(nsample), partial_aggregates, partial_aggregates);
	reduce_array<RSUM><<<gs(1,32*32),32*32,0,cs>>>(dim3(32*32), partial_aggregates, partial_aggregates);

	float result;
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(&result, partial_aggregates, sizeof(result), cudaMemcpyDeviceToHost));
	return result;
}
