#include <cmath>
#include "CudaMoGEMAfBindings.hpp"

#include "common/common.cuh"
#include "common/reduce.cuh"
#include "common/statistics.cuh"
#include "common/af.cuh"
#include "config.hpp"

#include <cublas_v2.h>

CudaMoGEMAfBindings::CudaMoGEMAfBindings(dim_t nfeature, dim_t nsample, dim_t nclass): CudaMoGEM(nfeature, nsample, nclass, NULL) {
	cs = afcu::getStream(af::getDevice());
	CUDA_DIE_UNLESS_SUCCESS(cublasCreate(&ch));
	CUDA_DIE_UNLESS_SUCCESS(cublasSetStream(ch, cs));
}

void CudaMoGEMAfBindings::af_cov_rowwise(af::array& Sigma, const af::array& samples, const af::array& mu) {
	af::array Var(samples.dims());
	float *dVar = Var.device<float>();
	float *dSigma = Sigma.device<float>();
	const float *dsamples = samples.device<float>();
	const float *dmu = mu.device<float>();

	cov_rowwise(ch, dSigma, dVar, dsamples, dim3_from_af_dim4(samples.dims()), dmu);

	mu.unlock(), samples.unlock(), Sigma.unlock(), Var.unlock();
}


void CudaMoGEMAfBindings::print_gm(const af::array& P, const af::array& mu, const af::array& Sigma) const {
    const float *dP = P.device<float>();
    const float *dmu = mu.device<float>();
    const float *dSigma = Sigma.device<float>();

    CudaMoGEM::print_gm(dP, dmu, dSigma);

    Sigma.unlock(), mu.unlock(), P.unlock();
}


void CudaMoGEMAfBindings::pdfs_from_samples(af::array& pdfs, const af::array& samples, const af::array& P, const af::array& mu, const af::array& Sigma) {
	float *dpdfs = pdfs.device<float>();
	const float *dsamples = samples.device<float>();
	const float *dP = P.device<float>();
	const float *dmu = mu.device<float>();

	af::array luSigma = Sigma.copy(), iSigma(Sigma.dims()), pSigma(nfeature,nclass,s32), batchStatus(nclass, s32);
	float *dluSigma = luSigma.device<float>();
	float *diSigma = iSigma.device<float>();
	int *dpSigma = pSigma.device<int>();
	int *dbatchStatus = batchStatus.device<int>();

	float **drefs;
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&drefs, 2*nclass*sizeof(*drefs)));
	CudaMoGEM::pdfs_from_samples(dpdfs, dsamples, dP, dmu, dluSigma, diSigma, dpSigma, dbatchStatus, drefs);

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(drefs));
	iSigma.unlock(), pSigma.unlock(), luSigma.unlock(), mu.unlock(), P.unlock(), samples.unlock(), pdfs.unlock();
}

void CudaMoGEMAfBindings::weights_from_pdfs(af::array& weights, const af::array& pdfs) {
	float *dweights = weights.device<float>();
	const float *dpdfs = pdfs.device<float>();

	CudaMoGEM::weights_from_pdfs(dweights, dpdfs);

	pdfs.unlock(), weights.unlock();
}
void CudaMoGEMAfBindings::classwise_share(af::array& P_with_respect_to_weights, const af::array &weights, float scalar) {
	af::array tmp_to_be_summed(nclass, std::ceil(nsample/1024.0)), ones = af::constant(1.0, weights.dims(1));
	const float *dones = ones.device<float>();
	float *dtmp_to_be_summed = tmp_to_be_summed.device<float>();
	const float *dweights = weights.device<float>();
	float *dshares = P_with_respect_to_weights.device<float>();

	CudaMoGEM::classwise_share(dshares, dweights, dones, dtmp_to_be_summed, scalar);

	P_with_respect_to_weights.unlock(), weights.unlock(), tmp_to_be_summed.unlock(), ones.unlock();
}
void CudaMoGEMAfBindings::mstep(af::array& P, af::array& mu, af::array& Sigma, const af::array& samples, const af::array& weights, af::array& FSC_A, af::array& FSC_B, const af::array& ones, const af::array& fallback_covariance) {
	float *dP = P.device<float>();
	float *dmu = mu.device<float>();
	float *dSigma = Sigma.device<float>();
	const float *dsamples = samples.device<float>();
	const float *dweights = weights.device<float>();

	af::array FSC_C(nfeature,nsample,nclass);
	float *dFSC_A = FSC_A.device<float>();
	float *dFSC_B = FSC_B.device<float>();
	float *dFSC_C = FSC_C.device<float>();
	const float *dones = ones.device<float>();
	const float *dfallback_covariance = fallback_covariance.device<float>();

	CudaMoGEM::mstep(dP, dmu, dSigma, dsamples, dweights, dFSC_A, dFSC_B, dFSC_C, dones, dfallback_covariance);

	fallback_covariance.unlock(), ones.unlock(), FSC_A.unlock(), FSC_B.unlock(), FSC_C.unlock();
	weights.unlock(), samples.unlock(), Sigma.unlock(), mu.unlock(), P.unlock();
}

void CudaMoGEMAfBindings::score(af::array& scores, const af::array& mu, const af::array& Sigma, const af::array& samples, const af::array& weights) {
	af::array P_with_respect_to_weights(nclass), P_to_be_summed(nclass, std::ceil(nsample/1024.0));
	float *dP_with_respect_to_weights = P_with_respect_to_weights.device<float>();
	float *dP_to_be_summed = P_to_be_summed.device<float>();

	float *dscores = scores.device<float>();
	const float *dmu = mu.device<float>();
	const float *dSigma = Sigma.device<float>();
	const float *dsamples = samples.device<float>();
	const float *dweights = weights.device<float>();

	af::array bin_bounds(NPROBABILITY_BINS+1, nfeature, nclass), weight_bins(NPROBABILITY_BINS,nfeature,nclass), ones = af::constant(1.0, (NPROBABILITY_BINS*nsample));
	float *dbin_bounds = bin_bounds.device<float>();
	float *dweight_bins = weight_bins.device<float>();
	const float *dones = ones.device<float>();

	CudaMoGEM::score(dscores, dmu, dSigma, dsamples, dweights, dP_with_respect_to_weights, dbin_bounds, dweight_bins, dP_to_be_summed, dones);

	weight_bins.unlock(), bin_bounds.unlock(), P_to_be_summed.unlock(), P_with_respect_to_weights.unlock(), ones.unlock();
	weights.unlock(), samples.unlock(), Sigma.unlock(), mu.unlock(), scores.unlock();
}

void CudaMoGEMAfBindings::determine_least_fitting_class(unsigned *hout_iworst_score, float *hout_worst_score, const af::array& scores) {
	const float *dscores = scores.device<float>();

	af::array index = af::constant(0.0, 1, u32), value = af::constant(0.0, 1);
	unsigned *dindex = index.device<unsigned>();
	float *dvalue = value.device<float>();

	CudaMoGEM::determine_least_fitting_class(hout_iworst_score, hout_worst_score, dscores, dindex, dvalue);

	value.unlock(), index.unlock(), scores.unlock();
}

void CudaMoGEMAfBindings::split_class(af::array& nP, af::array& nmu, af::array& nSigma, const af::array& P, const af::array& mu, const af::array& Sigma, unsigned iscore, float score) {
	float *dnP = nP.device<float>();
	float *dnmu = nmu.device<float>();
	float *dnSigma = nSigma.device<float>();
	const float *dP = P.device<float>();
	const float *dmu = mu.device<float>();
	const float *dSigma = Sigma.device<float>();

	CudaMoGEM::split_class(dnP, dnmu, dnSigma, dP, dmu, dSigma, iscore, score);

	nSigma.unlock(), nmu.unlock(), nP.unlock();
	Sigma.unlock(), mu.unlock(), P.unlock();
}

float CudaMoGEMAfBindings::log_likelihood(const af::array& pdfs) {
	const float *dpdfs = pdfs.device<float>();

	af::array partial_aggregates(nsample), ones = af::constant(1.0, pdfs.dims(1));
	float *dpartial_aggregates = partial_aggregates.device<float>();

	float result = CudaMoGEM::log_likelihood(dpdfs, dpartial_aggregates);

	partial_aggregates.unlock(), pdfs.unlock();
	return result;
}
