#include "CudaMoGEMStaticMemory.hpp"

#include <cassert>
#include <cstdio>
#include <utility>
#include <vector>

#include "common/common.cuh"
#include "common/reduce.cuh"
#include "common/statistics.cuh"
#include "common/matlike.hpp"
#include "config.hpp"

#include <cublas_v2.h>

CudaMoGEMStaticMemory::CudaMoGEMStaticMemory(unsigned nfeature, unsigned nsample, unsigned nclass, unsigned maxclass, cublasHandle_t ch):
	CudaMoGEM(nfeature, nsample, nclass, ch), maxclass(maxclass) {
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&ireduceIndex, sizeof(*ireduceIndex)));

	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&pSigma, maxclass*sizeof(*pSigma)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&batchStatus, maxclass*sizeof(*batchStatus)));

	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&drefs, maxclass*3*sizeof(*drefs)));

	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&P, maxclass*sizeof(*P)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpP, maxclass*sizeof(*tmpmu)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&mu, nfeature*maxclass*sizeof(*mu)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpmu, nfeature*maxclass*sizeof(*tmpmu)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&Sigma, nfeature*nfeature*maxclass*sizeof(*Sigma)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpSigma, nfeature*nfeature*maxclass*sizeof(*tmpSigma)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&luSigma, nfeature*nfeature*maxclass*sizeof(*luSigma)));

	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&fallback_covariance, nfeature*nfeature*sizeof(*fallback_covariance)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&pdfs, maxclass*nsample*sizeof(*pdfs)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&weights, maxclass*nsample*sizeof(*weights)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&scores, nfeature*maxclass*sizeof(*scores)));

	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&FSC_A, nfeature*nsample*maxclass*sizeof(*FSC_A)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&FSC_B, nfeature*nsample*maxclass*sizeof(*FSC_B)));
	CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&ones, nsample*sizeof(*ones)));
	set_array<<<gs(nsample/16,BS_1D),BS_1D,0,cs>>>(dim3(nsample), ones, 1.0F);
}

CudaMoGEMStaticMemory::~CudaMoGEMStaticMemory() {
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(ones));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(FSC_B));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(FSC_A));

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(scores));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(weights));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(pdfs));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(fallback_covariance));

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(luSigma));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpSigma));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(Sigma));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpmu));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(mu));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpP));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(P));

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(drefs));

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(batchStatus));
	CUDA_DIE_UNLESS_SUCCESS(cudaFree(pSigma));

	CUDA_DIE_UNLESS_SUCCESS(cudaFree(ireduceIndex));
}

void CudaMoGEMStaticMemory::MoGEM(const float *samples) {
	float log_likelihood_prev =0.0;
	for (size_t iteration=0; iteration < MAX_EM_ITERATIONS; iteration++) {
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(luSigma, Sigma, nfeature*nfeature*nclass*sizeof(*luSigma), cudaMemcpyDeviceToDevice));
		CudaMoGEM::pdfs_from_samples(pdfs, samples, P, mu, luSigma, tmpSigma, pSigma, batchStatus, drefs);
		CudaMoGEM::weights_from_pdfs(weights, pdfs);
		CudaMoGEM::mstep(P, mu, Sigma, samples, weights, FSC_A, FSC_B, ones, fallback_covariance);

		float log_likelihood = CudaMoGEM::log_likelihood(pdfs, FSC_A);
		if (std::abs((log_likelihood-log_likelihood_prev) / log_likelihood_prev) <= EM_EPSILON) {
			break;
		}
		log_likelihood_prev = log_likelihood;
	}
}

void CudaMoGEMStaticMemory::split_until_all_classes_within_confidence_interval(const float *samples) {
	CudaMoGEMStaticMemory::initilize_GMDistribution(P, mu, Sigma, samples, ones, FSC_A);
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(fallback_covariance, Sigma, nfeature*nfeature*nclass*sizeof(*fallback_covariance), cudaMemcpyDeviceToDevice));

	while(nclass < MAX_CLASSES) {
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(luSigma, Sigma, nfeature*nfeature*nclass*sizeof(*luSigma), cudaMemcpyDeviceToDevice));
		CudaMoGEM::pdfs_from_samples(pdfs, samples, P, mu, luSigma, tmpSigma, pSigma, batchStatus, drefs);
		CudaMoGEM::weights_from_pdfs(weights, pdfs);

		CudaMoGEM::score(scores, mu, Sigma, samples, weights, tmpP, FSC_A, FSC_B, ones);

		unsigned iworst_score;
		float worst_score;
		CudaMoGEM::determine_least_fitting_class(&iworst_score, &worst_score, scores, ireduceIndex, FSC_A);
		if (worst_score <= 1.0) {
			break;
		}
		CudaMoGEM::split_class(tmpP, tmpmu, tmpSigma, P, mu, Sigma, iworst_score, worst_score);
		std::swap(tmpP, P), std::swap(tmpmu, mu), std::swap(tmpSigma, Sigma);

		MoGEM(samples);
	}
}

void CudaMoGEMStaticMemory::print_gm() const {
	float *hP = (float *)alloca(nclass*sizeof(*hP));
	float *hmu = (float *)alloca(nfeature*nclass*sizeof(*hmu));
	float *hSigma = (float *)alloca(nfeature*nfeature*nclass*sizeof(*hSigma));

	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hP, P, nclass*sizeof(*hP), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hmu, mu, nfeature*nclass*sizeof(*hmu), cudaMemcpyDeviceToHost));
	CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hSigma, Sigma, nfeature*nfeature*nclass*sizeof(*hSigma), cudaMemcpyDeviceToHost));

	std::printf("%11s %11s %11s\n", "P", "mu", "Sigma");
	for (unsigned k=0; k < nclass; k++) {
		for (unsigned j=0; j<nfeature; j++) {
			if (j == 0) std::printf("%- 11.4g", hP[k]);
			else std::printf("%11s", "");

			std::printf(" %- 11.4g", hmu[k*nfeature + j]);
			for (unsigned i=0; i<nfeature; i++) {
				std::printf(" %- 11.4g", hSigma[k*nfeature*nfeature + j*nfeature + i]);
			}
			std::printf("\n");
		}
		std::printf("\n");
	}
}
