classdef Optimizer
    properties
        C;

        nclass;
        nfeature;
        nsample;

        P;
        mu;
        Sigma;
    end
    methods
        function [o] = split_until_all_classes_within_confidence_interval(o, samples)
            o.nclass = 1;
            [o.nfeature, o.nsample] = size(samples);

            o.P = ones(1, 1, 'gpuArray');
            o.mu = mean(samples, 2);
            o.Sigma = cov(samples');

            o.C = config();

            while o.nclass < o.C.MAX_CLASSES
				tsplit = toc;
                pdfs = pdfs_from_samples(o, samples);
                weights = weights_from_pdfs(o, pdfs);
                scores = score(o, weights, samples);

                [iclass,ifeature,worst_score] = determine_least_fitting_class(o, scores);
                if o.nclass >= o.C.MIN_CLASSES && worst_score <= 1.0
                    break
                end
                o = split_class(o, iclass, ifeature, worst_score);
                [o, niteration] = MoGEM(o, samples);
				fprintf('%-10s %10.6f %10.6f score %7.2f class %2d feature %d nMoGEMIteration %4lu\n', 'SPLIT', tsplit, toc, worst_score, iclass-1, ifeature-1, niteration);
            end
        end

        function [pdfs] = pdfs_from_samples(o, samples)
            pdfs = zeros(o.nclass, o.nsample, 'gpuArray');
            for i = 1:o.nclass
                mu_deviation = samples - repmat(o.mu(:,i), 1, o.nsample);
                qmd = sum(mu_deviation .* (o.Sigma(:,:,i)\mu_deviation), 1);

                ldS = 0.5*log(abs(det(o.Sigma(:,:,i))));
                lC = log(o.P(i)) - (o.nfeature/2)*log(2*pi);

                pdfs(i,:) = exp(lC+(qmd/-2)-ldS);
            end
            pdfs = pdfs .* isfinite(pdfs);
        end

        function [weights] = weights_from_pdfs(o, pdfs)
            samplewise_total = sum(pdfs,1);
            samplewise_total(find(not(isfinite(samplewise_total(:))))) = inf;
            weights = pdfs ./ repmat(samplewise_total, o.nclass, 1);
            weights(find(not(isfinite(weights(:))))) = 0;
        end

        function [scores] = score(o, weights, samples)
            cl = 1 - (1-o.C.CONFIDENCE_LEVEL)/(o.nfeature*o.nclass);
            threshold = chi2inv(cl, (o.C.NPROBABILITY_BINS - o.C.DOFS));

            classwise_share = sum(weights, 2);
				scores = zeros(o.nfeature, o.nclass, 'gpuArray');
            for j = 1:o.nclass
                expected_distribution = classwise_share(j)/o.C.NPROBABILITY_BINS;
                for i = 1:o.nfeature
                    edges = gpuArray(norminv((0:o.C.NPROBABILITY_BINS)/o.C.NPROBABILITY_BINS, o.mu(i,j), sqrt(o.Sigma(i,i,j))));
                    [~,~,bin_membership] = histcounts(samples(i,:), edges);
                    weight_bins = zeros(o.C.NPROBABILITY_BINS, 1, 'gpuArray');
                    for b = 1:o.C.NPROBABILITY_BINS
                        weight_bins(b) = sum((bin_membership==b) .* weights(j,:));
                    end
                    scores(i,j) = sum((weight_bins-expected_distribution).^2/expected_distribution) / threshold;
                end
            end
        end

        function [iclass, ifeature, worst_score] = determine_least_fitting_class(o, scores)
            [worst_score,iscore] = max(scores(:));
            iclass = floor((iscore-1) / o.nfeature) + 1;
            ifeature = mod(iscore-1, o.nfeature) + 1;
        end

        function [o] = split_class(o, iclass, ifeature, worst_score)
            o.nclass = o.nclass + 1;
            o.P(o.nclass) = o.P(iclass) / 2;
            o.P(iclass) = o.P(iclass) / 2;

            mean_shift = zeros(o.nfeature, 1, 'gpuArray');
            mean_shift(ifeature) = 0.5*sqrt(o.Sigma(ifeature,ifeature,iclass));
            o.mu(:,o.nclass) = o.mu(:,iclass) + mean_shift;
            o.mu(:,iclass) = o.mu(:,iclass) - mean_shift;

            o.Sigma(:,:,o.nclass) = o.Sigma(:,:,iclass);
        end

        function [o, i] = MoGEM(o, samples)
            log_likelihood_prev = 0.0;
            for i = 1:o.C.MAX_EM_ITERATIONS
                pdfs = pdfs_from_samples(o, samples);
                weights = weights_from_pdfs(o, pdfs);
                o = maximization(o, weights, samples);

                samplewise_total = sum(pdfs,1);
                log_likelihood = sum(log((samplewise_total>0) .* samplewise_total));
                if i >= o.C.MIN_EM_ITERATIONS && abs((log_likelihood_prev-log_likelihood) / log_likelihood) <= o.C.EM_EPSILON
                    break;
                end
                log_likelihood_prev = log_likelihood;
            end
        end

        function [o] = maximization(o, weights, samples)
            o.P = mean(weights,2);
            for i = 1:o.nclass
                weight_share = o.P(i) * o.nsample;
                if weight_share >= 2
                    w = repmat(weights(i,:), o.nfeature, 1);
                    o.mu(:,i) = sum(w.*samples,2) / weight_share;
                    mu_deviation = samples - repmat(o.mu(:,i), 1, o.nsample);
                    o.Sigma(:,:,i) = (mu_deviation * (w.*mu_deviation)') / weight_share;
                else
					fprintf('%-10s %10.6f %10.6f class %d\n', 'SINGULAR', 0, 0, i-1);
					o.P(i) = 1/o.nclass;
					o.mu(:,i) = samples(:,1+round(N-1)*rand);
					o.Sigma(:,:,i) = cov(samples') / o.nclass;
                end
            end
        end
    end
end
