#include <cfloat>
#include <cmath>
#include <cstdio>
#include <iostream>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <af/cuda.h>

#include "CudaMoGEMAfBindings.hpp"
#include "common/af_utils.hpp"
#include "common/monotonic.hpp"
#include "config.hpp"
#include "ext/prob/prob.hpp"

using namespace af;

class Optimizer {
private:
	unsigned nfeature, nsample;
	array samples, fallback_covariance;

	array pdfs_from_samples() const {
		array pdfs(nclass, nsample);
#ifdef OPTIMIZE_PDFS_FROM_SAMPLES
		CudaMoGEMAfBindings(nfeature,nsample,nclass).pdfs_from_samples(pdfs, samples, P, mu, Sigma);
#else
		array mu_deviation = tile(samples, 1, 1, nclass) - tile(moddims(mu, nfeature, 1, nclass), 1, nsample);
		float *hP = (float *)alloca(nclass * sizeof(*hP));
		P.host(hP);

		array qmd(nclass, nsample);
		float *hlC_minus_ldS = (float *)alloca(nclass * sizeof(*hlC_minus_ldS));
		for (unsigned i = 0; i < nclass; i++) {
			// NOTE: af::solve,af::inverse,... does not support batching
			//       and result from af::det is implicitly returned to host (i.e. no batch)
			qmd(i, span) = sum(mu_deviation(span,span,i)*solve(Sigma(span,span,i), mu_deviation(span, span, i)), 0);

			double ldS = 0.5*std::log(std::abs(det<float>(Sigma(span,span,i))));
			double lC = std::log(hP[i]) - (nfeature/2.0)*std::log(2*M_PI);
			hlC_minus_ldS[i] = lC-ldS;
		}

		array lC_minus_ldS = tile(array(nclass,hlC_minus_ldS), 1, nsample);
		pdfs = af_utils::replaceInf(exp(lC_minus_ldS + qmd/-2), 0.0);
#endif /*OPTIMIZE_PDFS_FROM_SAMPLES*/
		return pdfs;
	}

	array weights_from_pdfs(const array& pdfs) const {
		array weights(pdfs.dims());
#ifdef OPTIMIZE_WEIGHTS_FROM_PDFS
		CudaMoGEMAfBindings(nfeature,nsample,nclass).weights_from_pdfs(weights, pdfs);
#else
		array samplewise_total = sum(pdfs, 0);
		af_utils::update(samplewise_total, samplewise_total==0, INFINITY);

		weights = af_utils::replaceInf(pdfs / tile(samplewise_total, nclass), 0.0);
#endif /*OPTIMIZE_WEIGHTS_FROM_PDFS*/
		return weights;
	}

	void maximization(const array& weights) {
#ifdef OPTIMIZE_MAXIMIZATION
		array FSC_A(nfeature,nsample,nclass), FSC_B(nfeature,nsample,nclass), ones = constant(1.0, nsample);
		CudaMoGEMAfBindings(nfeature,nsample,nclass).mstep(P, mu, Sigma, samples, weights, FSC_A, FSC_B, ones, fallback_covariance);
#else
		P = mean(weights, 1);
		float *hP = (float *)alloca(nclass*sizeof(*hP));
		P.host(hP);
		for (unsigned i=0; i<nclass; i++) {
			float weight_share = hP[i]*nsample;
			if (weight_share >= 2) {
				af::array w = tile(weights(i,span), nfeature);

				mu(span,i) = sum(w*samples, 1) / weight_share;
				af::array mu_deviation = samples - tile(mu(span,i), 1, nsample);
				Sigma(span,span,i) = af::matmul(mu_deviation, w*mu_deviation, AF_MAT_NONE, AF_MAT_TRANS) / weight_share;
			} else {
				std::cout << i << "th class gone singular, randomizing" << std::endl;
				P(i) = 1.0/nclass;
				Sigma(span,span,i) = fallback_covariance;
				mu(span,i) = samples(span, std::rand()%nsample);
			}
		}
#endif /*OPTIMIZE_MAXIMIZATION*/
	}

	size_t MoGEM(void) {
		double log_likelihood_prev = 0.0;
		for (size_t iteration=1; iteration <= MAX_EM_ITERATIONS; iteration++) {
			array pdfs = pdfs_from_samples();
			maximization(weights_from_pdfs(pdfs));

			double log_likelihood;
#ifdef OPTIMIZE_LOG_LIKELIHOOD
			log_likelihood = CudaMoGEMAfBindings(nfeature,nsample,nclass).log_likelihood(pdfs);
#else
			array samplewise_total = sum(pdfs);
			log_likelihood = sum<double>(log((samplewise_total>0)*samplewise_total));
#endif /*OPTIMIZE_LOG_LIKELIHOOD*/
			if (iteration >= MIN_EM_ITERATIONS && std::abs((log_likelihood_prev-log_likelihood) / log_likelihood) <= EM_EPSILON) {
				return iteration;
			}
			log_likelihood_prev = log_likelihood;
		}
		return MAX_EM_ITERATIONS;
	}

	array score(const array& weights) const {
		array scores(nfeature, nclass);
#ifdef OPTIMIZE_SCORE
		CudaMoGEMAfBindings(nfeature,nsample,nclass).score(scores, mu, Sigma, samples, weights);
#else
		float *hmu = (float *)alloca(nfeature*nclass*sizeof(*hmu));
		mu.host(hmu);
		float *hsigma = (float *)alloca(nfeature*nclass*sizeof(*hsigma));
		sqrt(diag(Sigma)).host(hsigma);
		double cl = 1 - (1-CONFIDENCE_LEVEL)/(nfeature*nclass);
		double threshold = chi_square_cdf_inv(cl, NPROBABILITY_BINS-DOFS);

		array classwise_share = sum(weights, 1);
		for (unsigned j=0; j<nclass; j++) {
			array expected_distribution = tile(classwise_share(j)/NPROBABILITY_BINS, NPROBABILITY_BINS);
			array w = weights(j,span);
			for (unsigned i=0; i<nfeature; i++) {
				array X = samples(i, span), weight_bins(NPROBABILITY_BINS);
				for (unsigned bin=0; bin<NPROBABILITY_BINS; bin++) {
					double lower_bound = normal_cdf_inv((double)(bin) / NPROBABILITY_BINS, hmu[j*nfeature + i], hsigma[j*nfeature+i]);
					double upper_bound = normal_cdf_inv((double)(bin+1) / NPROBABILITY_BINS, hmu[j*nfeature + i], hsigma[j*nfeature+i]);
					array bin_filter = lower_bound<X && X<=upper_bound;
					weight_bins(bin) = sum(w*bin_filter);
				}
				scores(i, j) = sum<double>(pow(weight_bins-expected_distribution, 2)/expected_distribution)/threshold;
			}
		}
#endif /*OPTIMIZE_SCORE*/
		return scores;
	}

	void determine_least_fitting_class(unsigned *hout_iworst_score, float *hout_worst_score, const array& scores) {
#ifdef OPTIMIZE_DETERMINE_LEAST_FITTING_CLASS
		CudaMoGEMAfBindings(nfeature,0,nclass).determine_least_fitting_class(hout_iworst_score, hout_worst_score, scores);
#else
		max(hout_worst_score, hout_iworst_score, scores);
#endif /*OPTIMIZE_DETERMINE_LEAST_FITTING_CLASS*/
	}

	void split_class(unsigned iscore, float score) {
		array oldP = P, oldmu = mu, oldSigma = Sigma;
		P = array(nclass+1), mu = array(nfeature, nclass+1), Sigma = array(nfeature,nfeature, nclass+1);

#ifdef OPTIMIZE_SPLIT
		CudaMoGEMAfBindings(nfeature,nsample,nclass).split_class(P, mu, Sigma, oldP, oldmu, oldSigma, iscore, score);
#else
		unsigned iclass = iscore/nfeature, ifeature = iscore%nfeature;
		P(seq(nclass)) = oldP;
		P(iclass) /= 2;
		P(nclass) = P(iclass);

		mu(span, seq(nclass)) = oldmu;
		af::array mean_shift = constant(0.0, nfeature);
		mean_shift(ifeature) = 0.5*sqrt(oldSigma(ifeature,ifeature,iclass));
		mu(span, nclass) = oldmu(span,iclass) + mean_shift;
		mu(span, iclass) = oldmu(span,iclass) - mean_shift;

		Sigma(span, span, seq(nclass)) = oldSigma;
		Sigma(span, span, nclass) = oldSigma(span,span,iclass);
#endif /*OPTIMIZE_SPLIT*/
		nclass++;
	}
public:
	array P, mu, Sigma;
	unsigned nclass;

	Optimizer(array samples):
		nfeature(samples.dims(0)), nsample(samples.dims(1)), samples(samples), nclass(1) {
		P = constant(1.0, 1);
		mu = mean(samples, 1);
		Sigma = af_utils::cov_rowwise(samples, mu);
		fallback_covariance = Sigma;
	}

	void split_until_all_classes_within_confidence_interval(void) {
		while (nclass < MAX_CLASSES) {
			double tsplit = monotonic::since();
			array scores = score(weights_from_pdfs(pdfs_from_samples()));
			unsigned iworst_score;
			float worst_score;
			determine_least_fitting_class(&iworst_score, &worst_score, scores);
			if (nclass >= MIN_CLASSES && worst_score <= 1.0) {
				break;
			}
			split_class(iworst_score, worst_score);
			size_t niteration = MoGEM();
			std::printf("%-10s %10.6f %10.6f score %7.2f class %2d feature %d nMoGEMIteration %4lu\n", "SPLIT", tsplit, monotonic::since(), worst_score, iworst_score/nfeature, iworst_score%nfeature, niteration);
		}
	}
};

unsigned do_clustering(float *hP, float *hmu, float *hSigma, const float *hsamples, unsigned nfeature, unsigned nsample) {
	af::array samples(nfeature, nsample, hsamples);

	Optimizer optimizer(samples);
	optimizer.split_until_all_classes_within_confidence_interval();

	optimizer.P.host(hP);
	optimizer.mu.host(hmu);
	optimizer.Sigma.host(hSigma);

	return optimizer.nclass;
}
