#include <cstdio>
#include <cublas_v2.h>
#include <thrust/device_vector.h>

#include "CudaMoGEM.hpp"
#include "common/cuda.hpp"
#include "common/monotonic.hpp"
#include "config.hpp"

using namespace thrust;

#define PTR(dvector) raw_pointer_cast(dvector.data())

class MoGEMThrust: public CudaMoGEM {
private:
	device_vector<float> samples, ones, fallbackCovariance;

	void posteriors(device_vector<float>& pdfs, device_vector<float>& weights) {
		device_vector<float*> drefs(2*nclass);
		device_vector<float> iSigma(nfeature*nfeature*nclass), luSigma = Sigma;
		device_vector<int> pSigma(nfeature*nfeature*nclass), batchStatus(nclass);

		CudaMoGEM::pdfs_from_samples(PTR(pdfs), PTR(samples), PTR(P), PTR(mu), PTR(luSigma), PTR(iSigma), PTR(pSigma), PTR(batchStatus), PTR(drefs));
		CudaMoGEM::weights_from_pdfs(PTR(weights), PTR(pdfs));
	}
	size_t MoGEM() {
		float log_likelihood_prev = 0.0;
		for (unsigned iteration=1; iteration <= MAX_EM_ITERATIONS; iteration++) {
			device_vector<float> pdfs(nclass*nsample),weights(nclass*nsample);
			device_vector<float> FSC_A(nfeature*nsample*nclass), FSC_B(nfeature*nsample*nclass), FSC_C(nfeature*nsample*nclass);
			posteriors(pdfs, weights);

			CudaMoGEM::mstep(PTR(P), PTR(mu), PTR(Sigma), PTR(samples), PTR(weights), PTR(FSC_A), PTR(FSC_B), PTR(FSC_C), PTR(ones), PTR(fallbackCovariance));

			float log_likelihood = CudaMoGEM::log_likelihood(PTR(pdfs), PTR(FSC_A));
			if (iteration >= MIN_EM_ITERATIONS && std::abs((log_likelihood-log_likelihood_prev) / log_likelihood_prev) <= EM_EPSILON) {
				return iteration;
			}
			log_likelihood_prev = log_likelihood;
		}
		return MAX_EM_ITERATIONS;
	}
public:
	device_vector<float> P, mu, Sigma;

	MoGEMThrust(unsigned nfeature, unsigned nsample, unsigned nclass, const float *hsamples, cublasHandle_t ch=NULL):
		CudaMoGEM(nfeature, nsample, nclass, ch)
		, P(nclass), mu(nfeature*nclass), Sigma(nfeature*nfeature*nclass)
		, samples(hsamples, hsamples+nfeature*nsample)
   		, ones(nsample, 1.0) {
			device_vector<float> var(nfeature*nsample);
			CudaMoGEM::initilize_GMDistribution(PTR(P), PTR(mu), PTR(Sigma), PTR(samples), PTR(ones), PTR(var));
			fallbackCovariance = Sigma;
		}

	void split_until_all_classes_within_confidence_interval() {
		while (nclass < MAX_CLASSES) {
			double tsplit = monotonic::since();
			device_vector<float> pdfs(nclass*nsample), weights(nclass*nsample);
			posteriors(pdfs, weights);

			device_vector<float> scores(nfeature*nclass), P_with_respect_to_weights(nclass), P_to_be_summed(nfeature*(std::ceil(nsample/1024.0)));
			device_vector<float> bin_bounds((NPROBABILITY_BINS+1)*nfeature*nclass), weight_bins(NPROBABILITY_BINS*nfeature*nclass);
			CudaMoGEM::score(PTR(scores), PTR(mu), PTR(Sigma), PTR(samples), PTR(weights), PTR(P_with_respect_to_weights), PTR(bin_bounds), PTR(weight_bins), PTR(P_to_be_summed), PTR(ones));

			unsigned iworst_score;
			float worst_score;
			device_vector<unsigned> ireduceIndex(1);
			device_vector<float> ireduceValue(1);
			CudaMoGEM::determine_least_fitting_class(&iworst_score, &worst_score, PTR(scores), PTR(ireduceIndex), PTR(ireduceValue));
			if (nclass >= MIN_CLASSES && worst_score <= 1.0) {
				break;
			}
			device_vector<float> newP(nclass+1), newmu(nfeature*(nclass+1)), newSigma(nfeature*nfeature*(nclass+1));
			CudaMoGEM::split_class(PTR(newP), PTR(newmu), PTR(newSigma), PTR(P), PTR(mu), PTR(Sigma), iworst_score, worst_score);
			P.swap(newP), mu.swap(newmu), Sigma.swap(newSigma);

			size_t niteration = MoGEM();
			std::printf("%-10s %10.6f %10.6f score %7.2f class %2d feature %d nMoGEMIteration %4lu\n", "SPLIT", tsplit, monotonic::since(), worst_score, iworst_score/nfeature, iworst_score%nfeature, niteration);
		}
	}

	void extract_classifier(float *hP, float *hmu, float *hSigma) {
		copy(P.begin(), P.begin()+nclass, hP);
		copy(mu.begin(), mu.begin()+nfeature*nclass, hmu);
		copy(Sigma.begin(), Sigma.begin()+nfeature*nfeature*nclass, hSigma);
	}
};

unsigned do_clustering(float *hP, float *hmu, float *hSigma, const float *hsamples, unsigned nfeature, unsigned nsample) {
		cublasHandle_t ch;
		CUDA_DIE_UNLESS_SUCCESS(cublasCreate(&ch));

		MoGEMThrust moGEM(nfeature, nsample, 1, hsamples, ch);
		moGEM.split_until_all_classes_within_confidence_interval();
		moGEM.extract_classifier(hP, hmu, hSigma);

		return moGEM.nclass;
}
