#ifndef KERNELS_CUH
#define KERNELS_CUH
#include "common/blas.cuh"
#include "common/reduce.cuh"
#include "common/tile.cuh"
#include "config.hpp"

#if 1
// NOTE: cuBLAS LU format is L without implicit unit diagonal + U
template <typename T>
__device__ static inline void LUsolve(T *x, const T *LU, const int *pivot, const T *b, int n) {
	for (int i=0; i<n; i++) {
		x[i] = b[pivot[i]-1];
		for (int j=0; j<i; j++) {
			x[i] -= LU[j*n+i] * x[j];
		}
	}
	for (int i=n-1; i>=0; i--) {
		for (int j=i+1; j<n; j++) {
			x[i] -= LU[j*n+i] * x[j];
		}
		x[i] /= LU[i*n+i];
	}
}
template <typename T>
__device__ static inline T determinant(const T *LU, const int *pivot, int n) {
	T det = 1;
	bool negate = false;
	for (int i=0; i<n; i++) {
		det *= LU[i*n+i];
		negate ^= pivot[i] == i+1;
	}
	return pow(-1.0, (int)negate) * det;
}
__global__ void compute_pdfs(dim3 spdfs, dim3 ssamples, float *pdfs, const float *samples, const float *P, const float *mu, const int *pSigma, const float *luSigma) {
	int i = blockDim.x*blockIdx.x + threadIdx.x, j = blockDim.y*blockIdx.y + threadIdx.y;
	int nsample = spdfs.y, nclass = spdfs.x, nfeature = ssamples.x;
	if (j >= nsample || i >= nclass) { return; }

	const float *sample = samples + j*nfeature;
	mu += i*nfeature, pSigma += i*nfeature, luSigma += i*nfeature*nfeature;

	float mu_deviation[3];
	for (int k=0; k<nfeature; k++) {
		mu_deviation[k] = sample[k] - mu[k];
	}
	LUsolve(mu_deviation, luSigma, pSigma, mu_deviation, nfeature);
	float qmd = 0.0;
	for (int k=0; k<nfeature; k++) {
		qmd += (sample[k]-mu[k]) * mu_deviation[pSigma[k]-1];
	}

	float ldS = 0.5*log(abs(determinant(luSigma, pSigma, nfeature)));
	float lC = log(P[i]) - nfeature/2.0 * log(2*M_PI);

	float pdf = exp(lC + qmd/-2 - ldS);
	if (isinf(pdf)) { pdf = 0.0; }

	pdfs[j*nclass + i] = pdf;
}
#endif

extern "C" {__global__ void compute_pdfs(dim3 spdfs, dim3 ssamples, float *pdfs, const float *samples, const float *P, const float *mu, const int *pSigma, const float *luSigma, const float *iSigma) {
    int i = blockDim.x*blockIdx.x + threadIdx.x, j = blockDim.y*blockIdx.y + threadIdx.y;
    int nsample = spdfs.y, nclass = spdfs.x, nfeature = ssamples.x;
    if (j >= nsample || i >= nclass) { return; }

    const float *sample = samples + j*nfeature;
	mu += i*nfeature, pSigma += i*nfeature, luSigma += i*nfeature*nfeature, iSigma += i*nfeature*nfeature;

    float qmd = 0.0;
    float detSigma = 1;
    for (int ii=0; ii<nfeature; ii++) {
        detSigma *= pow(-1.0, (int)(pSigma[ii]==ii+1)) * luSigma[ii*nfeature+ii];

        float ith_matmul_iSigma_devj = 0.0;
        for (int jj=0; jj < nfeature; jj++) {
            ith_matmul_iSigma_devj += iSigma[jj*nfeature+ii] * (sample[jj]-mu[jj]);
        }
        qmd += (sample[ii]-mu[ii])*ith_matmul_iSigma_devj;
    }

    float ldS = 0.5*log(abs(detSigma));
    float lC = log(P[i]) - nfeature/2.0 * log(2*M_PI);

    float pdf = exp(lC + qmd/-2 - ldS);
    if (isinf(pdf)) { pdf = 0.0; }

    pdfs[j*nclass + i] = pdf;
}}

extern "C" {__global__ void samplewise_normalization(dim3 problem, float* out, const float* in) {
    unsigned i = blockDim.x*blockIdx.x + threadIdx.x;
    unsigned nsample = problem.y, nclass = problem.x;
    for (unsigned j = blockDim.y*blockIdx.y + threadIdx.y; j < nsample; j += blockDim.y*gridDim.y) {
        float value = i<nclass ? in[j*nclass+i] : 0.0;
        float total = _allreduce_warp<RSUM>(value, blockDim.x);
        if (i<nclass) {
            out[j*nclass + i] = value / total;
        }
    }
}}

extern "C" {__global__ void populate_covariance_factors(dim3 ssamples, dim3 sweights, float *mu_deviation_over_P, float *weighted_mu_deviation, const float *P, const float *mu, const float *samples, const float *weights) {
    unsigned i = blockDim.x*blockIdx.x + threadIdx.x;
    unsigned nsample = ssamples.y, nfeature = ssamples.x, nclass = sweights.x;
    for (unsigned j = blockDim.y*blockIdx.y + threadIdx.y; j < nsample*nclass; j += blockDim.y*gridDim.y) {
        unsigned c=j/nsample, x=j%nsample;

        float mu_deviation = samples[x*nfeature + i] - mu[c*nfeature + i];
        mu_deviation_over_P[c*nsample*nfeature + x*nfeature + i] = mu_deviation / (nsample*P[c]);
        weighted_mu_deviation[c*nsample*nfeature + x*nfeature + i] = weights[x*nclass + c] * mu_deviation;
    }
}}

extern "C" {__global__ void weight_vectorized_samples(dim3 ssamples, dim3 sweights, float *weighted_samples, const float *samples, const float *weights, const float *P) {
	unsigned i = blockDim.x*blockIdx.x + threadIdx.x;
	unsigned nsample = ssamples.y, nfeature = ssamples.x, nclass = sweights.x;
	for (unsigned j = blockDim.y*blockIdx.y + threadIdx.y; j < nsample*nclass; j += blockDim.y*gridDim.y) {
		unsigned c=j/nsample, x=j%nsample;

		float w = weights[x*nclass + c], s = samples[x*nfeature + i];
		weighted_samples[x*nclass*nfeature + c*nfeature + i] = (w*s) / (nsample*P[c]);
	}
}}

extern "C" {__global__ void ksplit_class(dim3 snSigma, float *nP, float *nmu, float *nSigma, const float *P, const float *mu, const float *Sigma, unsigned class_to_split, unsigned feature_to_shift) {
	unsigned i = blockDim.x*blockIdx.x + threadIdx.x, j = blockDim.y*blockIdx.y + threadIdx.y, c = blockDim.z*blockIdx.z + threadIdx.z;
	unsigned nfeature = snSigma.x, nclass = snSigma.z;
	if (i >= nfeature || j >= nfeature || c >= nclass) return;

	float mean_shift = 0.0, mean_shift_value = 0.5*sqrt(Sigma[class_to_split*nfeature*nfeature + feature_to_shift*nfeature + feature_to_shift]);
	float P_factor = 1.0;
	if (c == nclass-1 || c == class_to_split) {
		P_factor = 1.0/2.0;
		if (i == feature_to_shift) {
			mean_shift = c==class_to_split ? -mean_shift_value : mean_shift_value;
		}
	}
	unsigned c_in = c==nclass-1 ? class_to_split : c;

	nSigma[c*nfeature*nfeature + j*nfeature + i] = Sigma[c_in*nfeature*nfeature + j*nfeature + i];
	if (j == 0) {
		nmu[c*nfeature + i] = mu[c_in*nfeature + i] + mean_shift;
		if (i == 0) {
			nP[c] = P_factor * P[c_in];
		}
	}
}}

extern "C" {__global__ void weight_binning(dim3 problem, dim3 sweights, float *weight_bins, const float *bin_bounds, const float *samples, const float *weights) {
	unsigned b = blockDim.x*blockIdx.x + threadIdx.x, y = blockDim.y*blockIdx.y + threadIdx.y;
	unsigned nsample = sweights.y, nfeature = problem.y, nclass = sweights.x;

	for (unsigned i = blockDim.z*blockIdx.z + threadIdx.z; i < nsample*nclass;) {
		unsigned c=i/nsample, x=i%nsample;

		float lower_bound = bin_bounds[c*nfeature*(NPROBABILITY_BINS+1) + y*(NPROBABILITY_BINS+1) + b];
		float upper_bound = bin_bounds[c*nfeature*(NPROBABILITY_BINS+1) + y*(NPROBABILITY_BINS+1) + b+1];

		float bin = 0.0;
		for (; x<nsample; x += blockDim.z*gridDim.z) {
			float sample = samples[x*nfeature + y];
			if (lower_bound<sample && sample<=upper_bound) {
				bin += weights[x*nclass+c];
			}
		}
		if (bin > 0.0) {
			atomicAdd(weight_bins + c*nfeature*NPROBABILITY_BINS + y*NPROBABILITY_BINS + b, bin);
		}

		i = c*nsample + x;
	}
}}

extern "C" {__global__ void scores_from_weight_bins(dim3 problem, float *scores, const float *weight_bins, const float *P, float threshold, unsigned nsample) {
	unsigned b = blockDim.x*blockIdx.x + threadIdx.x, i = blockDim.y*blockIdx.y + threadIdx.y, c = blockDim.z*blockIdx.z + threadIdx.z;
	unsigned nfeature = problem.y;

	float expected_distribution_share = (P[c]*nsample) / NPROBABILITY_BINS;
	float error = 0.0;
	if (b < NPROBABILITY_BINS) {
		float bin = weight_bins[c*nfeature*NPROBABILITY_BINS + i*NPROBABILITY_BINS + b];
		error = pow(bin-expected_distribution_share, 2);
	}

	error = _reduce_to_head_of_warp<RSUM>(error, blockDim.x);
	if (b == 0) {
		scores[c*nfeature + i] = error / (threshold*expected_distribution_share);
	}
}}

extern "C" {__global__ void compute_log_likelihood(dim3 problem, float* log_likelihood_to_be_summed, const float* pdfs) {
	unsigned i = blockDim.x*blockIdx.x + threadIdx.x;
	unsigned nsample = problem.y, nclass = problem.x;
	for (unsigned j = blockDim.y*blockIdx.y + threadIdx.y; j < nsample; j += blockDim.y*gridDim.y) {
		float value = i<nclass ? pdfs[j*nclass+i] : 0.0;
		float likelihood = _reduce_to_head_of_warp<RSUM>(value, blockDim.x);
		if (i==0) {
			log_likelihood_to_be_summed[j] = likelihood==0 ? 0 : log(likelihood);
		}
	}
}}

#endif /*KERNELS_CUH*/
