#include <cmath>
#include <cstdio>
#include <vector>

#include "CudaMoGEM.hpp"
#include "common/common.cuh"
#include "common/monotonic.hpp"
#include "config.hpp"

class Optimizer: public CudaMoGEM {
private:
	unsigned *ireduce_index;
	int *pSigma, *batch_status;
	float **drefs;

	float *tmpP, *tmpmu, *tmpSigma;
	float *fallback_covariance, *pdfs, *weights, *scores;
	float *FSC_A, *FSC_B, *FSC_C, *ones;

	size_t MoGEM() {
		float log_likelihood_prev =0.0;
		for (size_t iteration=1; iteration <= MAX_EM_ITERATIONS; iteration++) {
			CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(tmpSigma, Sigma, nfeature*nfeature*nclass*sizeof(*tmpSigma), cudaMemcpyDeviceToDevice));
			CudaMoGEM::pdfs_from_samples(pdfs, samples, P, mu, tmpSigma, FSC_A, pSigma, batch_status, drefs);
			CudaMoGEM::weights_from_pdfs(weights, pdfs);
			CudaMoGEM::mstep(P, mu, Sigma, samples, weights, FSC_A, FSC_B, FSC_C, ones, fallback_covariance);

			float log_likelihood = CudaMoGEM::log_likelihood(pdfs, FSC_A);
			if (iteration >= MIN_EM_ITERATIONS && std::abs((log_likelihood-log_likelihood_prev) / log_likelihood_prev) <= EM_EPSILON) {
				return iteration;
			}
			log_likelihood_prev = log_likelihood;
		}
		return MAX_EM_ITERATIONS;
	}
public:
	float *P, *mu, *Sigma;
	float *samples;

	Optimizer(unsigned nfeature, unsigned nsample, const float *hsamples, cublasHandle_t ch=NULL): CudaMoGEM(nfeature, nsample, 1, ch) {
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&ireduce_index, sizeof(*ireduce_index)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&pSigma, MAX_CLASSES*sizeof(*pSigma)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&batch_status, MAX_CLASSES*sizeof(*batch_status)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&drefs, MAX_CLASSES*3*sizeof(*drefs)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpP, MAX_CLASSES*sizeof(*tmpP)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpmu, nfeature*MAX_CLASSES*sizeof(*tmpmu)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&tmpSigma, nfeature*nfeature*MAX_CLASSES*sizeof(*tmpSigma)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&fallback_covariance, nfeature*nfeature*sizeof(*fallback_covariance)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&pdfs, MAX_CLASSES*nsample*sizeof(*pdfs)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&weights, MAX_CLASSES*nsample*sizeof(*weights)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&scores, nfeature*MAX_CLASSES*sizeof(*scores)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&FSC_A, nfeature*nsample*MAX_CLASSES*sizeof(*FSC_A)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&FSC_B, nfeature*nsample*MAX_CLASSES*sizeof(*FSC_B)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&FSC_C, nfeature*nsample*MAX_CLASSES*sizeof(*FSC_C)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&ones, nsample*sizeof(*ones)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&P, MAX_CLASSES*sizeof(*P)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&mu, nfeature*MAX_CLASSES*sizeof(*mu)));
		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&Sigma, nfeature*nfeature*MAX_CLASSES*sizeof(*Sigma)));

		CUDA_DIE_UNLESS_SUCCESS(cudaMalloc(&samples, nfeature*nsample*sizeof(*samples)));
		CUDA_DIE_UNLESS_SUCCESS(cudaThreadSynchronize());

		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(samples, hsamples, nfeature*nsample*sizeof(*samples), cudaMemcpyHostToDevice));
		set_array<<<gs(nsample/16,BS_1D),BS_1D,0,cs>>>(dim3(nsample), ones, 1.0F);
	}
	~Optimizer() {
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(samples));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(Sigma));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(mu));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(P));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(ones));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(FSC_C));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(FSC_B));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(FSC_A));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(scores));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(weights));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(pdfs));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(fallback_covariance));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpSigma));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpmu));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(tmpP));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(drefs));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(batch_status));
		CUDA_DIE_UNLESS_SUCCESS(cudaFree(pSigma));

		CUDA_DIE_UNLESS_SUCCESS(cudaFree(ireduce_index));
	}

	void split_until_all_classes_within_confidence_interval() {
		CudaMoGEM::initilize_GMDistribution(P, mu, Sigma, samples, ones, FSC_A);
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(fallback_covariance, Sigma, nfeature*nfeature*nclass*sizeof(*fallback_covariance), cudaMemcpyDeviceToDevice));

		while(nclass < MAX_CLASSES) {
			double tsplit = monotonic::since();
			CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(tmpSigma, Sigma, nfeature*nfeature*nclass*sizeof(*tmpSigma), cudaMemcpyDeviceToDevice));
			CudaMoGEM::pdfs_from_samples(pdfs, samples, P, mu, tmpSigma, FSC_A, pSigma, batch_status, drefs);
			CudaMoGEM::weights_from_pdfs(weights, pdfs);

			CudaMoGEM::score(scores, mu, Sigma, samples, weights, tmpP, FSC_A, FSC_B, FSC_C, ones);

			unsigned iworst_score;
			float worst_score;
			CudaMoGEM::determine_least_fitting_class(&iworst_score, &worst_score, scores, ireduce_index, FSC_A);
			if (nclass >= MIN_CLASSES && worst_score <= 1.0) {
				break;
			}
			CudaMoGEM::split_class(tmpP, tmpmu, tmpSigma, P, mu, Sigma, iworst_score, worst_score);
			std::swap(tmpP, P), std::swap(tmpmu, mu), std::swap(tmpSigma, Sigma);

			size_t niteration = MoGEM();
			std::printf("%-10s %10.6f %10.6f score %7.2f class %2d feature %d nMoGEMIteration %4lu\n", "SPLIT", tsplit, monotonic::since(), worst_score, iworst_score/nfeature, iworst_score%nfeature, niteration);
		}
	}
};

unsigned do_clustering(float *hP, float *hmu, float *hSigma, const float *hsamples, unsigned nfeature, unsigned nsample) {
		cublasHandle_t ch;
		CUDA_DIE_UNLESS_SUCCESS(cublasCreate(&ch));

		Optimizer optimizer(nfeature, nsample, hsamples, ch);
		optimizer.split_until_all_classes_within_confidence_interval();

		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hP, optimizer.P, optimizer.nclass*sizeof(*hP), cudaMemcpyDeviceToHost));
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hmu, optimizer.mu, nfeature*optimizer.nclass*sizeof(*hmu), cudaMemcpyDeviceToHost));
		CUDA_DIE_UNLESS_SUCCESS(cudaMemcpy(hSigma, optimizer.Sigma, nfeature*nfeature*optimizer.nclass*sizeof(*hSigma), cudaMemcpyDeviceToHost));

		return optimizer.nclass;
}
