#!/usr/bin/env python3
from common import cmx
from config import *
from ctypes import c_void_p
from functools import reduce
from random import randint
from sys import argv
from scipy import stats
from time import time

import array as ha
import arrayfire as af
import math as m

tstart = time()

def cov_rowwise(X, mu):
    Var = X-af.tile(mu, 1,X.dims()[1])
    return af.matmul(Var,Var,af.MATPROP.NONE,af.MATPROP.TRANS) / X.dims()[1]


def replace(A, pattern, filler):
    if af.any_true(pattern):
        print("replacing %d with %g"%(af.count(pattern), filler))
        A[pattern] = filler

    return A


def replaceInf(A, filler):
    return replace(A, af.isinf(A), filler)


class Optimizer(object):
    def __init__(self, samples):
        self.samples = samples

        self.nclass = 1
        self.nfeature = samples.dims()[0]
        self.nsample = samples.dims()[1]

        self.P = af.constant(1.0, 1)
        self.mu = af.mean(samples, dim=1)
        self.Sigma = cov_rowwise(samples,self.mu)
        self._fallback_covariance = self.Sigma


    def _print_gm(self):
        print("P:", self.P)
        print("mu:", self.mu)
        print("Sigma", self.Sigma)


    def _pdfs_from_samples(self):
        hP = self.P.to_array()
        pdfs = af.Array(dims=(self.nclass, self.nsample))
        for i in range(0,self.nclass):
            mu_deviation = self.samples-af.tile(self.mu[:,i], 1, self.nsample)
            qmd = af.sum(mu_deviation*af.solve(self.Sigma[:,:,i], mu_deviation), 0)

            ldS = 0.5*m.log(abs(af.det(self.Sigma[:,:,i])))
            lC = m.log(hP[i]) - (self.nfeature/2.0)*m.log(2*m.pi)

            pdfs[i,:] = af.exp(lC + qmd/-2 - ldS)

        return replaceInf(pdfs, 0.0)


    def _weights_from_pdfs(self, pdfs):
        samplewise_total = af.sum(pdfs, 0)
        samplewise_total = replace(samplewise_total, samplewise_total==0, float('inf'))
        return replaceInf(pdfs / af.tile(samplewise_total, self.nclass), 0.0)


    def _maximization(self, weights):
        self.P = af.mean(weights, dim=1)
        P = self.P.to_array()
        for i in range(0, self.nclass):
            weight_share = P[i] * self.nsample
            if weight_share >= 2:
                w = af.tile(weights[i,:], self.nfeature)
                self.mu[:,i] = af.sum(w*samples, 1) / weight_share
                mu_deviation = samples - af.tile(self.mu[:,i], 1, self.nsample)
                self.Sigma[:,:,i] = af.matmul(mu_deviation, w*mu_deviation, af.MATPROP.NONE, af.MATPROP.TRANS) / weight_share
            else:
                print("%-10s %10.6f %10.6f class %d"%(0, 0, i))
                self.P[i] = 1.0/self.nclass;
                self.mu[:,i] = self.samples[:,randint(0,nsample-1)]
                self.Sigma[:,:,i] = self._fallback_covariance / self.nclass


    def _log_likelihood(self, pdfs):
        samplewise_total = af.sum(pdfs,0)
        return af.sum(af.log((samplewise_total>0)*samplewise_total))


    def _MoGEM(self):
        log_likelihood_prev = 0.0
        for i in range(1, MAX_EM_ITERATIONS+1):
            pdfs = self._pdfs_from_samples()
            weights = self._weights_from_pdfs(pdfs)
            self._maximization(weights)

            log_likelihood = self._log_likelihood(pdfs)
            if i >= MIN_EM_ITERATIONS and abs((log_likelihood_prev-log_likelihood)/log_likelihood) <= EM_EPSILON:
                return i

            log_likelihood_prev = log_likelihood
        return MAX_EM_ITERATIONS


    def _score(self, weights):
        hmu = self.mu.to_array()
        hsigma = af.sqrt(af.diag(self.Sigma)).to_array()

        cl = 1 - (1-CONFIDENCE_LEVEL)/(self.nfeature*self.nclass)
        threshold = stats.chi2.ppf(cl, NPROBABILITY_BINS-DOFS)

        classwise_share = af.sum(weights, 1)
        scores = af.Array(dims=(self.nfeature,self.nclass))
        for j in range(0,self.nclass):
            expected_distribution = af.tile(classwise_share[j]/NPROBABILITY_BINS,NPROBABILITY_BINS)
            w = weights[j,:]

            for i in range(0, self.nfeature):
                X = self.samples[i,:]
                weight_bins = af.Array(dims=(NPROBABILITY_BINS,1))

                for b in range(0, NPROBABILITY_BINS):
                    lower_bound = stats.norm.ppf(b/NPROBABILITY_BINS, hmu[j*self.nfeature+i], hsigma[j*self.nfeature+i])
                    upper_bound = stats.norm.ppf((b+1)/NPROBABILITY_BINS, hmu[j*self.nfeature+i], hsigma[j*self.nfeature+i])
                    binned = (X>lower_bound) * (X<=upper_bound)
                    weight_bins[b] = af.sum(w*binned,1)

                scores[i,j] = af.sum(af.pow(weight_bins-expected_distribution, 2)/expected_distribution)/threshold

        return scores


    def _determine_least_fitting_class(self, scores):
        return af.imax(scores)


    def _split_class(self, iscore, score):
        iclass = iscore/self.nfeature; ifeature = iscore%self.nfeature

        oldP = self.P; oldmu = self.mu; oldSigma = self.Sigma

        self.nclass += 1
        self.P = af.Array(dims=(self.nclass,1)); self.mu = af.Array(dims=(self.nfeature,self.nclass)); self.Sigma = af.Array(dims=(self.nfeature,self.nfeature,self.nclass))

        self.P[0:-1] = oldP
        self.P[iclass] /= 2
        self.P[-1] = self.P[iclass]

        self.mu[:,0:-1] = oldmu
        mean_shift = af.constant(0.0, self.nfeature)
        mean_shift[ifeature] = 0.5*af.sqrt(oldSigma[ifeature,ifeature,iclass])
        self.mu[:,-1] = self.mu[:,iclass] + mean_shift
        self.mu[:,iclass] -= mean_shift

        self.Sigma[:,:,0:-1] = oldSigma
        self.Sigma[:,:,-1] = oldSigma[:,:,iclass]


    def split_until_all_classes_within_confidence_interval(self):
        while self.nclass < MAX_CLASSES:
            tsplit = time()
            pdfs = self._pdfs_from_samples()
            weights = self._weights_from_pdfs(pdfs)
            scores = self._score(weights)

            worst_score, iworst_score = self._determine_least_fitting_class(scores)
            if self.nclass >= MIN_CLASSES and worst_score <= 1.0:
                break

            self._split_class(iworst_score, worst_score)
            niteration = self._MoGEM()
            print("%-10s %10.6f %10.6f score %7.2f class %2d feature %d nMoGEMIteration %4lu"%("SPLIT", tsplit-tstart, time()-tstart, worst_score, iworst_score/nfeature, iworst_score%nfeature, niteration));

nfeature = len(argv)-1
(t,dims) = cmx.parse(argv[1])
nsample_full = reduce(lambda x,y:x*y, dims, 1)
nsample = int(nsample_full/SUBSAMPLE)

samples = af.Array(dims=(nfeature,nsample))
for i in range(0, nfeature):
    with open(argv[i+1], 'rb') as f:
        a = ha.array('f')
        a.fromfile(f, nsample_full)
        samples[i,:] = af.Array(a[slice(0,nsample_full,SUBSAMPLE)], (1,nsample))

tAMoG = time()
o = Optimizer(samples)
o.split_until_all_classes_within_confidence_interval()
print("%-10s %10.6f %10.6f"%("AMOG", tAMoG-tstart, time()-tstart));

cmx.write(o.P, "classifier_P")
cmx.write(o.mu, "classifier_mu")
cmx.write(o.Sigma, "classifier_Sigma")
