#include "af_geotiff.hpp"
#include "af_utils.hpp"

#include <cassert>

#include <fcntl.h>
#include <libgeotiff/geotiffio.h>
#include <libgeotiff/xtiffio.h>
#include <sys/mman.h>
#include <unistd.h>

static inline af::dim4 dims_from_tiff_metadata(uint32_t h, uint32_t w, uint16_t spp, uint16_t sampleformat) {
	bool is_complex = sampleformat == SAMPLEFORMAT_COMPLEXIEEEFP || sampleformat == SAMPLEFORMAT_COMPLEXINT;
	return af::dim4(is_complex?2:1, spp, w, h);
}
static inline af::dtype dtype_from_tiff_metadata(uint16_t bps, uint16_t spp, uint16_t sampleformat) {
	af::dim4 dims = dims_from_tiff_metadata(1,1,spp,sampleformat);
	switch (sampleformat) {
		case SAMPLEFORMAT_IEEEFP:
		case SAMPLEFORMAT_COMPLEXIEEEFP:
			if (bps/dims[0] == 32) return f32;
			else if (bps/dims[0] == 64) return f64;
			else assert("INVALID FLOAT" == NULL);
		case SAMPLEFORMAT_UINT:
			if (bps/dims[0] == 8) return u8;
			else if (bps/dims[0] == 16) return u16;
			else if (bps/dims[0] == 32) return u32;
			else if (bps/dims[0] == 64) return u64;
			else assert("INVALID UNSIGNED INTEGER" == NULL);
		case SAMPLEFORMAT_INT:
		case SAMPLEFORMAT_COMPLEXINT:
			if (bps/dims[0] == 8) return b8;
			else if (bps/dims[0] == 16) return s16;
			else if (bps/dims[0] == 32) return s32;
			else if (bps/dims[0] == 64) return s64;
			else assert("INVALID SIGNED INTEGER" == NULL);
		default:
			assert("INVALID TYPE" == NULL);

	}
	assert("NOT REACHABLE" == NULL);
	return f32;
}

af::array af_geotiff::load(const char *filename) {
	int fd = open(filename, O_RDONLY);
	assert(fd >= 0);
	TIFF *tif = XTIFFFdOpen(fd, filename, "rC");
	assert(tif != NULL);

	uint32_t w, h;
	uint16_t bps, compression, photometric, planarconfig, sampleformat, spp;
	assert(TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w) == 1);
	assert(TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h) == 1);
	assert(TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bps) == 1);

	TIFFGetFieldDefaulted(tif, TIFFTAG_COMPRESSION, &compression);
	TIFFGetFieldDefaulted(tif, TIFFTAG_PHOTOMETRIC, &photometric);
	TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planarconfig);
	TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLEFORMAT, &sampleformat);
	TIFFGetFieldDefaulted(tif, TIFFTAG_SAMPLESPERPIXEL, &spp);

	assert(compression == COMPRESSION_NONE);
	assert(photometric == PHOTOMETRIC_MINISBLACK);
	assert(planarconfig == PLANARCONFIG_CONTIG);

	af::array A(dims_from_tiff_metadata(h, w, spp, sampleformat), dtype_from_tiff_metadata(bps, spp, sampleformat));

	uint64_t *stripOffsets, *stripSize;
	assert(TIFFGetField(tif, TIFFTAG_STRIPOFFSETS, &stripOffsets) == 1);
	assert(TIFFGetField(tif, TIFFTAG_STRIPBYTECOUNTS, &stripSize) == 1);

	for (int i=1; i < A.dims(3); i++) {
		// ensure continuous pixel array
		assert(stripSize[i] == stripSize[0]);
		assert(stripOffsets[i]-stripOffsets[i-1] == stripSize[0]);
	}

	char *content = (char *)mmap(NULL, stripOffsets[0] + A.bytes(), PROT_READ, MAP_PRIVATE, fd, 0);
	assert(content != MAP_FAILED);

	af_utils::writeDuckTyped(A, content+stripOffsets[0], A.bytes());

	assert(munmap(content, stripOffsets[0] + A.bytes()) == 0);

	XTIFFClose(tif); // NOTE: closes fd internally

	return af::reorder(A, 3, 2, 1, 0);
}
af::array af_geotiff::load(const std::string &filename) {
	return af_geotiff::load(filename.c_str());
}
