#include "af_utils.hpp"

#include <cassert>
#include <cstdint>
#include <cstdlib>
#include <cstring>
#include <ostream>

#include <arrayfire.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>


using namespace af;

const char *af_utils::dtype_names[] = {"f32", "c32", "f64", "c64", "b8", "s32", "u32", "u8", "s64", "u64", "s16", "u16"};

bool af_utils::dim4_div(const dim4 &dividend, const dim4 &divisor, dim4 *quotient, dim4 *remainder) {
	std::lldiv_t a = std::lldiv(dividend[0], divisor[0]);
	std::lldiv_t b = std::lldiv(dividend[1], divisor[1]);
	std::lldiv_t c = std::lldiv(dividend[2], divisor[2]);
	std::lldiv_t d = std::lldiv(dividend[3], divisor[3]);

	if (quotient != NULL) {
		*quotient = dim4(a.quot, b.quot, c.quot, d.quot);
	}
	if (remainder != NULL) {
		*remainder = dim4(a.rem, b.rem, c.rem, d.rem);
	}

	return a.rem==0 && b.rem==0 && c.rem==0 && d.rem==0; // i.e. divides
}
bool af_utils::dim4_le(const dim4 &a, const dim4 &b) {
	return a[0]<=b[0] && a[1]<=b[1] && a[2]<=b[2] && a[3]<=a[3];
}

af::dim4 af_utils::dim4_parse(const char *decimalnumbers, dim_t filler) {
	assert(decimalnumbers != NULL);

	dim4 dims(filler, filler, filler, filler);
	const char *nstr = decimalnumbers;
	for (int i=0; i<5; i++) {
		assert(i < 4);

		dims[i] = std::strtoull(nstr, (char **)&nstr, 10);

		assert(nstr != NULL);
		if (*nstr == '\0') {
			break;
		} else {
			nstr++; // skip some separator
		}
	}
	return dims;
}
af::dim4 af_utils::dim4_parse(const std::string &decimalnumbers, dim_t filler) {
	return af_utils::dim4_parse(decimalnumbers.c_str(), filler);
}

af::array af_utils::crop(const af::array &A, const Geometry &g) {
	assert(g.dims[0]+g.offset[0] <= A.dims(0));
	assert(g.dims[1]+g.offset[1] <= A.dims(1));
	assert(g.dims[2]+g.offset[2] <= A.dims(2));
	assert(g.dims[3]+g.offset[3] <= A.dims(3));

	af::seq cropper[4] = {af::span};
	for (int i=0; i<4; i++) {
		cropper[i] = af::seq(g.offset[i], g.offset[i]+g.dims[i]-1);
	}

	return A(cropper[0], cropper[1], cropper[2], cropper[3]);
}
array af_utils::cropToDivide(const array &in, const dim4 &divisor) {
	assert(dim4_le(divisor, in.dims()));

	dim4 quotient;
	dim4_div(in.dims(), divisor, &quotient, NULL);

	return af_utils::crop(in, af_utils::Geometry(divisor*quotient, af::dim4(0,0,0,0)));
}

af::array af_utils::maximizeRange(const af::array &img) {
	af::array out = img.as(f32);

	af::array m = min(min(out,1)), M = max(max(out,1));
	af::array range = M-m;

	out -= tile(m, out.dims(0), out.dims(1));
	out /= tile(range, out.dims(0), out.dims(1));

	return out;
}
af::array af_utils::normalize(const af::array &A, double new_min, double new_max) {
	if (new_max-new_min == 0.0) {
		return constant(new_max, A.dims(), A.type());
	}

	double m = af::min<double>(A);
	double M = af::max<double>(A);

	if (M-m == 0.0) {
		// default case results in division by 0 which yields infinity
		// this is not perfect, but best approach I could come up with
		// though maybe throw exception would be more appropriate
		return constant((new_max-new_min) / 2, A.dims(), A.type());
	}

	return ((A-m) * (new_max-new_min)/(M-m)) + new_min;
}
af::array af_utils::clampAbove(const af::array &A, double upperThreshold) {
	return A*(A<=upperThreshold) + upperThreshold*(A>upperThreshold);
}
af::array af_utils::clampBelow(const af::array &A, double lowerThreshold) {
	if (lowerThreshold < 0.0 && !af_utils::issigned(A)) {
		return A;
	}
	return A*(A>=lowerThreshold) + lowerThreshold*(A<lowerThreshold);
}
af::array af_utils::clamp(const af::array &A, double lowerThreshold, double upperThreshold) {
	if (lowerThreshold < 0.0 && !af_utils::issigned(A)) {
		return af_utils::clampAbove(A, upperThreshold);
	}
	return A*(lowerThreshold<=A && A<=upperThreshold) + lowerThreshold*(A<lowerThreshold) + upperThreshold*(A>upperThreshold);
}
af::array af_utils::clampStatistically(const af::array &A, double sigmaBelow, double sigmaAbove) {
	if (isnan(sigmaBelow) && isnan(sigmaAbove)) {
		return A;
	}
	double u = af::mean<double>(A);
	double s = af::stdev<double>(A);
	double lowerThreshold = u + sigmaBelow*s, upperThreshold = u + sigmaAbove*s;

	if (isnan(sigmaBelow)) {
		return af_utils::clampAbove(A, upperThreshold);
	} else if (isnan(sigmaAbove)) {
		return af_utils::clampBelow(A, lowerThreshold);
	}
	return af_utils::clamp(A, lowerThreshold, upperThreshold);
}
void af_utils::update(af::array &A, const af::array &filter, double filler, dim_t *filterCount) {
    dim_t _filterCount = filter.nonzeros();
    if (_filterCount != 0) {
        A(filter) = filler;
    } // else: af::array(ZEROES) throws exception
    if (filterCount != NULL) {
        *filterCount = _filterCount;
    }
}
af::array af_utils::replace(const af::array &A, const af::array &filter, double filler, dim_t *filterCount) {
    af::array ret = A;
    af_utils::update(ret, filter, filler, filterCount);
    return ret;
}
af::array af_utils::replaceNaN(const af::array &A, double filler, dim_t *nNaN) {
    af::array ret = A;
    if (A.isfloating()) { // otherwise NaN is not defined
        af::array NaNs = af::isNaN(A);
        af_utils::update(ret, NaNs, filler, nNaN);
    } else if (nNaN != NULL) {
        *nNaN = 0;
    }
    return ret;
}
af::array af_utils::replaceInf(const af::array &A, double filler, dim_t *nInf) {
    af::array ret = A;
    if (A.isfloating()) { // otherwise NaN is not defined
        af::array infs = af::isInf(A);
        af_utils::update(ret, infs, filler, nInf);
    } else if (nInf != NULL) {
        *nInf = 0;
    }
    return ret;
}
af::array af_utils::nonInfNaN(const af::array &A, double filler, dim_t *nInf, dim_t *nNaN) {
    return af_utils::replaceNaN(af_utils::replaceInf(A, filler, nInf), filler, nNaN);
}
af::dim4 af_utils::squeeze(const af::dim4& dims) {
	af::dim4 new_dims(1);
	for (int i_old=0, i_new=0; i_old <4; i_old++) {
		if (dims[i_old] != 1) {
			new_dims[i_new++] = dims[i_old];
		}
	}
	return new_dims;
}
af::array af_utils::squeeze(const af::array &A) {

	return af::moddims(A, af_utils::squeeze(A.dims()));
}

void af_utils::writeDuckTyped(af::array &A, void *ptr, const size_t bytes, af::source src) {
	switch(A.type()) {
		case f32: A.write((float *)ptr, bytes, src); break;
		case c32: A.write((cfloat *)ptr, bytes, src); break;
		case f64: A.write((double *)ptr, bytes, src); break;
		case c64: A.write((cdouble *)ptr, bytes, src); break;
		case b8:  A.write((char *)ptr, bytes, src); break;
		case s32: A.write((int32_t *)ptr, bytes, src); break;
		case u32: A.write((uint32_t *)ptr, bytes, src); break;
		case u8:  A.write((uint8_t *)ptr, bytes, src); break;
		case s64: A.write((long long *)ptr, bytes, src); break;
		case u64: A.write((unsigned long long *)ptr, bytes, src); break;
		case s16: A.write((int16_t *)ptr, bytes, src); break;
		case u16: A.write((uint16_t *)ptr, bytes, src); break;
		default: assert("NOT IMPLEMENTED" == NULL); break;
	}
}
void af_utils::hostDuckTyped(const af::array &A, void *ptr, const size_t bytes) {
	assert(bytes >= A.bytes());
	switch(A.type()) {
		case f32: A.host((float *)ptr); break;
		case c32: A.host((cfloat *)ptr); break;
		case f64: A.host((double *)ptr); break;
		case c64: A.host((cdouble *)ptr); break;
		case b8:  A.host((char *)ptr); break;
		case s32: A.host((int32_t *)ptr); break;
		case u32: A.host((uint32_t *)ptr); break;
		case u8:  A.host((uint8_t *)ptr); break;
		case s64: A.host((long long *)ptr); break;
		case u64: A.host((unsigned long long *)ptr); break;
		case s16: A.host((int16_t *)ptr); break;
		case u16: A.host((uint16_t *)ptr); break;
		default: assert("NOT IMPLEMENTED" == NULL); break;
	}
}

af::array af_utils::load_cmx(const char *filename) {
	af::dim4 dims = af_utils::dims_from_cmx(filename);
	af::dtype type = af_utils::dtype_from_cmx(filename);

	af::array A(dims, type);

	int fd = open(filename, O_RDONLY);
	assert(fd >= 0);
	void *content = mmap(NULL, A.bytes(), PROT_READ, MAP_PRIVATE, fd, 0);
	assert(content != MAP_FAILED);

	af_utils::writeDuckTyped(A, content, A.bytes());

	assert(munmap(content, A.elements()) == 0);
	assert(close(fd) == 0);

	return A;
}
af::array af_utils::load_cmx(const std::string &filename) {
	return af_utils::load_cmx(filename.c_str());
}
void af_utils::store_cmx(const af::array &A, const char *filename_without_cmx_suffix, int oflag, mode_t mode) {
	std::string filename = filename_without_cmx_suffix + af_utils::cmx_suffix(A);

	int fd = oflag&O_CREAT ? open(filename.c_str(), oflag, mode) : open(filename.c_str(), oflag);
	assert(fd >= 0);
	lseek(fd, A.bytes()-1, SEEK_SET);
	char a_byte = 0;
	write(fd, &a_byte, 1);
	void *content = mmap(NULL, A.bytes(), PROT_WRITE, MAP_SHARED, fd, 0);
	assert(content != MAP_FAILED);

	af_utils::hostDuckTyped(A, content, A.bytes());

	assert(munmap(content, A.elements()) == 0);
	assert(close(fd) == 0);
}
void af_utils::store_cmx(const af::array &A, const std::string &filename_without_cmx_suffix, int oflag, mode_t mode) {
	af_utils::store_cmx(A, filename_without_cmx_suffix.c_str(), oflag, mode);
}

af::dtype af_utils::dtype_parse(const char *str, size_t n) {
	for (size_t i=0; i < sizeof(af_utils::dtype_names)/sizeof(*af_utils::dtype_names); i++) {
		if (strncmp(af_utils::dtype_names[i], str, n) == 0) {
			return (af::dtype)i;
		}
	}
	return (af::dtype)-1;
}
af::dtype af_utils::dtype_parse(const std::string &str) {
	return af_utils::dtype_parse(str.c_str(), str.length());
}

static inline void cmx_sanity_check(const char *filename, const char **type, const char **dims) {
	const char *suffix = std::strrchr(filename, '.');
	assert(suffix != NULL);
	const char *dims_suffix = std::strchr(suffix, '_');
	assert(dims_suffix != NULL);

	assert(dims_suffix-suffix == sizeof(".cmXX")-1 || dims_suffix-suffix == sizeof(".cmXXX")-1);
	assert(std::strncmp(suffix, ".cm", 3) == 0);

	if (type != NULL) {
		*type = suffix + sizeof(".cm")-1;
	}
	if (dims != NULL) {
		*dims = dims_suffix;
	}
}
af::dtype af_utils::dtype_from_cmx(const char *filename) {
	const char *type, *dims;
	cmx_sanity_check(filename, &type, &dims);
	return af_utils::dtype_parse(type, dims-type);
}
af::dtype af_utils::dtype_from_cmx(const std::string &filename) {
	return af_utils::dtype_from_cmx(filename.c_str());
}
af::dim4 af_utils::dims_from_cmx(const char *filename) {
	const char *dims_suffix;
	cmx_sanity_check(filename, NULL, &dims_suffix);

	dim4 dims(1);
	for(int nth_dim = 0; *dims_suffix !='\0'; nth_dim++) {
		assert(nth_dim < 4);
		const char *without_separator = dims_suffix+1;
		dims[nth_dim] = std::strtoull(without_separator, (char **)&dims_suffix, 10);
	}

	return dims;
}
af::dim4 af_utils::dims_from_cmx(const std::string &filename) {
	return af_utils::dims_from_cmx(filename.c_str());
}
std::string af_utils::cmx_suffix(af::dtype type, const af::dim4 &dims) {
	std::string suffix = ".cm";
	suffix += af_utils::dtype_names[type];

	suffix += "_" + std::to_string(dims[0]);
	int required_dims = (dims[1]!=1) | (dims[2]!=1)<<1 | (dims[3]!=1)<<2;
	if (required_dims >= 1<<0) { suffix += "x" + std::to_string(dims[1]); }
	if (required_dims >= 1<<1) { suffix += "x" + std::to_string(dims[2]); }
	if (required_dims >= 1<<2) { suffix += "x" + std::to_string(dims[3]); }

	return suffix;
}
std::string af_utils::cmx_suffix(const af::array &A) {
	return af_utils::cmx_suffix(A.type(), A.dims());
}

af::array af_utils::cov_rowwise(const af::array &A, const af::array &A_rowwise_mu) {
    af::array Var = A-af::tile(A_rowwise_mu, 1, A.dims(1));
    return af::matmul(Var, Var, AF_MAT_NONE, AF_MAT_TRANS) / A.dims(1);
}
af::array af_utils::cov_rowwise(const af::array &A) {
	return af_utils::cov_rowwise(A, af::mean(A, 1));
}
af::array af_utils::cov_colwise(const af::array &A, const af::array &A_colwise_mu) {
    af::array Var = A-af::tile(A_colwise_mu, A.dims(0));
    return af::matmul(Var, Var, AF_MAT_TRANS, AF_MAT_NONE) / A.dims(0);
}
af::array af_utils::cov_colwise(const af::array &A) {
	return af_utils::cov_rowwise(A, af::mean(A, 0));
}
af::array af_utils::clog(const af::array &Z) {
	// using MATLAB's definition, see https://se.mathworks.com/help/matlab/ref/log.html
	return log(abs(Z)) + cfloat(0,1)*atan2(imag(Z), real(Z));
}

bool af_utils::issigned(const af::array &A) {
	switch (A.type()) {
		case u8:
		case u16:
		case u32:
		case u64:
			return false;
		default:
			return true;
	}
}
bool af_utils::isfloat(const af::array &A) {
	switch (A.type()) {
		case c32:
		case c64:
		case f32:
		case f64:
			return true;
		default:
			return false;
	}
}

static std::string stats_line_complex(const af::array &A) {
	char buf[128];
	cdouble m = min<cdouble>(A);
	cdouble M = max<cdouble>(A);
	cdouble u = mean<cdouble>(A);
	cdouble s2 = var<cdouble>(A);

	snprintf(buf, sizeof(buf), "%10llu %e+%ei %e+%ei %e+%ei %e+%ei", A.elements(), real(m),imag(m), real(M),imag(M), real(u),imag(u), real(s2),imag(s2));

	return std::string(buf);
}
static std::string stats_line_real(const af::array &A) {
	char buf[128];
	double m = min<double>(A);
	double M = max<double>(A);
	double u = mean<double>(A);
	double s2 = var<double>(A);

	snprintf(buf, sizeof(buf), "%10llu %e %e %e %e", A.elements(), m, M, u, s2);

	return std::string(buf);
}
void af_utils::dump_info(std::ostream &out, const af::array &A, const std::string &name) {
	std::string (*formatter)(const af::array &);
	if (A.iscomplex()) {
		formatter = stats_line_complex;
	} else {
		formatter = stats_line_real;
	}

	dim_t nInf, nNaN;
	af::array sanitized = af_utils::nonInfNaN(A, 0.0, &nInf, &nNaN);

	out << name << ": (#inf: "<<(nInf)<<", #NaN:"<<(nNaN)<<") " << std::endl;
	out << "\telements: " << formatter(sanitized) << std::endl;
	for (int i=0;i<4;i++) {
		dim_t this_dim = A.dims(i);
		if (this_dim <= 1) {
			continue;
		}

		out << "\tmean("<<i<<"):  " <<formatter(mean(sanitized, i)) << std::endl;
	}
}
