#pragma once
#include "reduce.cuh"

template<bool A_TRANSPOSED, bool B_TRANSPOSED, typename T>
__device__ void gemm1(dim3 problem, T *C_to_be_summed, const T *A, const T *B) {
    unsigned m = problem.x, n = problem.y, p = problem.z;
    for (unsigned i=0; i<m; i++) {
        for (unsigned j=0; j<p; j++) {
            T acc = 0.0;
			unsigned first_k = blockDim.x*blockIdx.x + threadIdx.x;
            for (unsigned k = first_k; k < n; k += blockDim.x*gridDim.x) {
				T a = A_TRANSPOSED? A[i*m+k] : A[k*m+i];
				T b = B_TRANSPOSED? B[k*p+j] : B[j*p+k];
                acc += a*b;
            }
            __syncthreads();
            acc = _reduce_block<RSUM>(acc, sqrt((T)blockDim.x));
            if (threadIdx.x == 0) {
                C_to_be_summed[j*m*gridDim.x + i*gridDim.x + first_k/blockDim.x] = acc;
            }
        }
    }
}
extern "C" {__global__ void Sgemm1_nn(dim3 problem, float *C_to_be_summed, const float *A, const float *B) {
	gemm1<false,false>(problem, C_to_be_summed, A, B);
}}
extern "C" {__global__ void Sgemm1_nt(dim3 problem, float *C_to_be_summed, const float *A, const float *B) {
	gemm1<false,true>(problem, C_to_be_summed, A, B);
}}
extern "C" {__global__ void Sgemm1_tn(dim3 problem, float *C_to_be_summed, const float *A, const float *B) {
	gemm1<true,false>(problem, C_to_be_summed, A, B);
}}
extern "C" {__global__ void Sgemm1_tt(dim3 problem, float *C_to_be_summed, const float *A, const float *B) {
	gemm1<true,true>(problem, C_to_be_summed, A, B);
}}

template<typename T>
__device__ void gemm2(dim3 problem, T beta, T *C, T alpha, const T *C_to_be_summed) {
    unsigned m = problem.x, n_pass_2 = problem.y, p = problem.z;
    for (unsigned i=0; i<m; i++) {
        for (unsigned j=0; j<p; j++) {
            T v = threadIdx.x<n_pass_2 ? C_to_be_summed[j*m*n_pass_2 + i*n_pass_2 + threadIdx.x] : 0.0;
            v = _reduce_block<RSUM>(v, sqrt((T)blockDim.x));
            if (threadIdx.x == 0) {
				T previous_C = beta==0.0 ? 0.0 : beta*C[j*m+i];
                C[j*m + i] = previous_C + alpha*v;
            }
        }
    }
}
extern "C" {__global__ void Sgemm2(dim3 problem, float beta, float *C, float alpha, const float *C_to_be_summed) {
	gemm2(problem, beta, C, alpha, C_to_be_summed);
}}

static inline void sum_rowwise(cudaStream_t cs, float *out, const float *in, dim3 in_span, const float *ones, float *out_to_be_summed, float scalar=1.0) {
	dim3 ts(32*32);
	unsigned bs = gs(dim3(in_span.y),ts).x;
	Sgemm1_nn<<<bs,ts,0,cs>>>(in_span, out_to_be_summed, in, ones);
	in_span.y = bs;
	Sgemm2<<<1,ts_warp_aligned(bs),0,cs>>>(in_span, 0.0, out, scalar, out_to_be_summed);
}
