#pragma once

#include "common.cuh"
#include <cublas_v2.h>

enum reducer {
	RSUM,
};
template <enum reducer op, typename T> struct Reducer {
	__device__ T init();
	__device__ T operator() (T rhs, T lhs);
};
template <typename T> struct Reducer<RSUM,T> {
	__device__ T init() {return 0;}
	__device__ T operator() (T rhs, T lhs) { return rhs+lhs; }
};

template <enum reducer reducer, typename T>
__device__ static inline T _reduce_threadwise(const T* in, unsigned ldin, unsigned n) {
	struct Reducer<reducer,T> reduce;
	T accumulator = *in;
	for (int i = 1; i<n; i++) {
		accumulator = reduce(accumulator, in[ldin*i]);
	}
	return accumulator;
}
template <enum reducer reducer, typename T>
__device__ static inline T _reduce_to_head_of_warp(T value, int width=32) {
	struct Reducer<reducer,T> reduce;
	for (int offset = width/2; offset > 0; offset>>=1) {
		value = reduce(value, __shfl_down(value, offset, width));
	}
	return value;
}
template <enum reducer reducer, typename T>
__device__ static inline T _reduce_block(T v, int width=32) {
	static __shared__ T blockwise_v_out[32];
	unsigned lane = threadIdx.x%width, wid = threadIdx.x/width;

	v = _reduce_to_head_of_warp<reducer,T>(v, width);
	if (lane == 0) blockwise_v_out[wid] = v;
	__syncthreads();

	if (wid == 0) v = blockwise_v_out[lane];
	v = _reduce_to_head_of_warp<reducer,T>(v, width);
	__syncthreads();
	return v;
}
template <enum reducer reducer, typename T>
__global__ static void reduce_array(dim3 sin, T *value, const T *in) {
	struct Reducer<reducer,T> reduce;
	unsigned i = blockIdx.x*blockDim.x + threadIdx.x;
	T accumulator = _reduce_block<reducer,T>(i<sin.x ? in[i] : reduce.init(), sqrt((float)blockDim.x));
	if (threadIdx.x == 0) {
		value[i/blockDim.x] = accumulator;
	}
}
template <enum reducer reducer, typename T>
__device__ static inline T _allreduce_warp(T value, int width=32) {
	struct Reducer<reducer,T> reduce;
	for (int offset = width/2; offset > 0; offset>>=1) {
		value = reduce(value, __shfl_xor(value, offset, width));
	}
	return value;
}

enum ireducer {
	IRMAX,
	IRMIN,
};
template <enum ireducer op, typename T> struct IReducer {
	__device__ T init();
	__device__ void operator() (unsigned lhi, T lhv, unsigned rhi, T rhv, unsigned *i, T *v);
};
template <typename T> struct IReducer<IRMAX,T> {
	__device__ T init() {return -INFINITY;}
	__device__ void operator() (unsigned lhi, T lhv, unsigned rhi, T rhv, unsigned *i, T *v) {
		if (lhv >= rhv) *i = lhi, *v = lhv;
		else *i = rhi, *v = rhv;
	}
};
template <typename T> struct IReducer<IRMIN,T> {
	__device__ T init() {return INFINITY;}
	__device__ void operator() (unsigned rhi, T rhv, unsigned lhi, T lhv, unsigned *i, T *v) {
		if (lhv <= rhv) *i = lhi, *v = lhv;
		else *i = rhi, *v = rhv;
	}
};

template <enum ireducer ireducer, typename T>
__device__ static inline void _ireduce_warp(unsigned i, T v, unsigned *i_out, T *v_out, int width=32) {
	struct IReducer<ireducer,T> ireduce;
	for (int offset = width/2; offset > 0; offset>>=1) {
		T other = __shfl_down(v, offset);
		unsigned other_i = __shfl_down(i, offset);
		ireduce(i,v, other_i,other, &i,&v);
	}
	*i_out = i, *v_out = v;
}
template <enum ireducer ireducer, typename T>
__device__ static inline void _ireduce_block(unsigned i, T v, unsigned *i_out, T *v_out, int width=32) {
	static __shared__ T blockwise_v_out[32];
	static __shared__ unsigned blockwise_i_out[32];
	unsigned lane = threadIdx.x%width, wid = threadIdx.x/width;

	_ireduce_warp<ireducer,T>(i, v, &i, &v, width);
	if (lane == 0) blockwise_i_out[wid] = i, blockwise_v_out[wid] = v;
	__syncthreads();

	if (wid == 0) i = blockwise_i_out[lane], v = blockwise_v_out[lane];
	_ireduce_warp<ireducer,T>(i, v, &i, &v, width);
	__syncthreads();

	if (threadIdx.x == 0) {
		*i_out = i, *v_out = v;
	}
}
template <enum ireducer ireducer, typename T>
__global__ static void ireduce_array(dim3 sin, unsigned *index, T *value, const T *in) {
	struct IReducer<ireducer,T> ireduce;
	unsigned i = blockIdx.x*blockDim.x + threadIdx.x;
	unsigned max_i;
	T max;
	_ireduce_block<ireducer,T>(i, i<sin.x ? in[i] : ireduce.init(), &max_i, &max, sqrt((float)blockDim.x));
	if (threadIdx.x == 0) {
		index[i/blockDim.x] = max_i;
		value[i/blockDim.x] = max;
	}
}

static inline unsigned ts_warp_aligned(unsigned n) {
	for (unsigned log2_sqrt_ts = 1; log2_sqrt_ts <= 5; log2_sqrt_ts <<= 1) { // maximum ts := 1024 = (2^5)^2
		unsigned ts = (1<<log2_sqrt_ts) * (1<<log2_sqrt_ts);
		if (n <= ts) {
			return ts;
		}
	}
	return 0;
}

static inline void sum_rowwise(cublasHandle_t cublas, float *out, const float *in, dim3 in_span, const float *ones, float scalar=1.0, int ones_stride=1, int out_stride=1) {
	float beta=0.0;
	CUDA_DIE_UNLESS_SUCCESS(cublasSgemv(cublas, CUBLAS_OP_N, in_span.x, in_span.y, &scalar, in, in_span.x, ones, ones_stride, &beta, out, out_stride));
}
