#include <cassert>
#include <cstring>
#include <iostream>
#include <string>

#include <endian.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

enum direction { TO_HOST, FROM_HOST, };
enum endian { LITTLE, BIG, };

enum direction direction = TO_HOST;
enum endian endian = BIG;
size_t wordsize = 4;

char *me;
static void die(const std::string &error = "") {
	if (!error.empty()) {
		std::cerr << "Error: " << error << std::endl;
	}
	std::cout << "Usage: "<<me<<"[-f?rom host?] [-t?o host] [-b?ig] [-l?ittle] [-n?etwork] [-s WORDSIZE] <IN> <OUT>" << std::endl;
	std::exit(error.empty() ? EXIT_SUCCESS : EXIT_FAILURE);
}

template <typename T, enum direction D, enum endian E> inline T reorder_word(T);
template <> inline uint16_t reorder_word<uint16_t,FROM_HOST,BIG>(uint16_t w) { return htobe16(w); }
template <> inline uint16_t reorder_word<uint16_t,FROM_HOST,LITTLE>(uint16_t w) { return htole16(w); }
template <> inline uint16_t reorder_word<uint16_t,TO_HOST,BIG>(uint16_t w) { return be16toh(w); }
template <> inline uint16_t reorder_word<uint16_t,TO_HOST,LITTLE>(uint16_t w) { return le16toh(w); }
template <> inline uint32_t reorder_word<uint32_t,FROM_HOST,BIG>(uint32_t w) { return htobe32(w); }
template <> inline uint32_t reorder_word<uint32_t,FROM_HOST,LITTLE>(uint32_t w) { return htole32(w); }
template <> inline uint32_t reorder_word<uint32_t,TO_HOST,BIG>(uint32_t w) { return be32toh(w); }
template <> inline uint32_t reorder_word<uint32_t,TO_HOST,LITTLE>(uint32_t w) { return le32toh(w); }
template <> inline uint64_t reorder_word<uint64_t,FROM_HOST,BIG>(uint64_t w) { return htobe64(w); }
template <> inline uint64_t reorder_word<uint64_t,FROM_HOST,LITTLE>(uint64_t w) { return htole64(w); }
template <> inline uint64_t reorder_word<uint64_t,TO_HOST,BIG>(uint64_t w) { return be64toh(w); }
template <> inline uint64_t reorder_word<uint64_t,TO_HOST,LITTLE>(uint64_t w) { return le64toh(w); }

template <typename T, enum direction D, enum endian E> inline void reorder_data(void *in, void *out, size_t n) {
	for (size_t i = 0; i < n; i++) {
		((T*)out)[i] = reorder_word<T,D,E>(((T*)in)[i]);
	}
}

template <typename T> inline void swap_endian(void *in, void *out, size_t nbyte, enum direction d, enum endian e) {
	assert(nbyte % sizeof(T) == 0);
	size_t nword = nbyte/sizeof(T);

	if      (d==FROM_HOST && e==BIG)    reorder_data<T,FROM_HOST,BIG>(in, out, nword);
	else if (d==FROM_HOST && e==LITTLE) reorder_data<T,FROM_HOST,LITTLE>(in, out, nword);
	else if (d==TO_HOST   && e==BIG)    reorder_data<T,TO_HOST,BIG>(in, out, nword);
	else if (d==TO_HOST   && e==LITTLE) reorder_data<T,TO_HOST,LITTLE>(in, out, nword);
	else assert("UNREACHABLE");
}

int main(int argc, char *argv[]) {
	me = argv[0];

	while (--argc >= 0 && **(++argv) == '-') {
		switch(argv[0][1]) {
			case 'f': direction = FROM_HOST; break;
			case 't': direction = TO_HOST; break;

			case 'b': endian = BIG; break;
			case 'l': endian = LITTLE; break;
			case 'n': endian = LITTLE; break;

			case 'h':
			case '?':
				die();

			default:
				assert(argc >= 2);
				switch(argv[0][1]) {
					case 's': wordsize = strtoull(*(--argc,++argv), NULL, 10); break;
					default:
							 die("Invalid option: \""+std::string(argv[0])+"\"");
				}
		}
	}
	if (argc != 2) die("Require 2 arguments");

	int fdIn = open(argv[0], O_RDONLY), fdOut = open(argv[1], O_RDWR|O_TRUNC|O_CREAT, 0666);
	assert(fdIn >= 0); assert(fdOut >= 0);

	off_t nbyte = lseek(fdIn, 0, SEEK_END);
	lseek(fdOut, nbyte-1, SEEK_SET);
	char a_byte = 0;
	write(fdOut, &a_byte, 1);

	void *in = mmap(NULL, nbyte, PROT_READ, MAP_SHARED, fdIn, 0);
	void *out = mmap(NULL, nbyte, PROT_WRITE, MAP_SHARED, fdOut, 0);

	switch (wordsize) {
		case 2: swap_endian<uint16_t>(in, out, nbyte, direction, endian); break;
		case 4: swap_endian<uint32_t>(in, out, nbyte, direction, endian); break;
		case 8: swap_endian<uint32_t>(in, out, nbyte, direction, endian); break;
		default: die("Invalid wordsize "+std::to_string(wordsize));
	}

	assert(munmap(in, nbyte) == 0); assert(munmap(out, nbyte) == 0);
	assert(close(fdIn) == 0); assert(close(fdOut) == 0);
	return 0;
}
