#include <cassert>
#include <cstring>
#include <string>
#include <arrayfire.h>

#include "../af_utils.hpp"

enum normalization {
	NONE,
	TOGETHER,
	PER_COLOR,
};

enum normalization normalization = TOGETHER;
double output_scalar_x = 1.0, output_scalar_y = 1.0;
double clamp_lower_sigma = NAN, clamp_upper_sigma = 3.0;
std::string output_filename = "";
af_cspace_t colorspace = AF_RGB;

char *me;
static void die(const std::string &error = "") {
if (!error.empty()) {
	std::cerr << "Error: " << error << std::endl;
}
std::cout << "Usage: "<<me<<"[-H?SV?] [-R?GB?] [-Y?CbCr?] [-P?er color normalization?] [-l LOWER_SIGMA_CLAMP] -[u UPPER_SIGMA_CLAMP] [-o OUTPUT] [-x OUTPUT_SCALAR_X] [-y OUTPUT_SCALAR_Y] <GREY/RED/RGB|constant> [<GREEN|constant>] [<BLUE|constant>]" << std::endl;
	std::exit(error.empty() ? EXIT_SUCCESS : EXIT_FAILURE);
}

static inline af::array input_transformation(const af::array &A) {
	af::array ret = A;
	if (A.iscomplex()) {
		ret = abs(A);
	}
	if (output_scalar_x != 1.0 && output_scalar_y != 1.0) {
		ret = af::scale(ret, output_scalar_y, output_scalar_x);
	}
	if (normalization == PER_COLOR) {
		ret = af_utils::normalize(af_utils::clampStatistically(af_utils::nonInfNaN(ret), clamp_lower_sigma, clamp_upper_sigma));
	}
	return ret;
}

int main(int argc, char *argv[]) {
	me = argv[0];
	af::setBackend(AF_BACKEND_CPU); // lots of RAM

	while (--argc >= 0 && **(++argv) == '-') {
		switch(argv[0][1]) {
			case 'H': colorspace = AF_HSV; break;
			case 'R': colorspace = AF_RGB; break;
			case 'Y': colorspace = AF_YCbCr; break;

			case 'P': normalization = PER_COLOR; break;

			case 'h':
			case '?':
				die();

			default:
				assert(argc >= 2);
				switch(argv[0][1]) {
					case 'l': clamp_lower_sigma = strtod(*(--argc,++argv), NULL); break;
					case 'o': output_filename = std::string(*(--argc,++argv)); break;
					case 'x': output_scalar_x = strtod(*(--argc,++argv), NULL); break;
					case 'y': output_scalar_y = strtod(*(--argc,++argv), NULL); break;
					case 'u': clamp_upper_sigma = strtod(*(--argc,++argv), NULL); break;

					default:
							 die("Invalid option: \""+std::string(argv[0])+"\"");
				}
		}
	}
	if (argc < 1) die("Require arguments");

	int first_non_empty_argument;
	// TODO check that argument is not constant (usually as Hue in HSV, so not that important)
	for (first_non_empty_argument=0; *argv[first_non_empty_argument] == '\0'; first_non_empty_argument++) {
		assert(argc > first_non_empty_argument);
	}
	char *output_candidate = argv[first_non_empty_argument];
	af::array img = input_transformation(af_utils::load_cmx(output_candidate));

	assert(img.dims(3) == 1);
	switch(img.dims(2)) {
		case 3: assert(argc == 1); // fallthrough
		case 1: break;
		default:
			die("First valid matrix has invalid number of color slices");
	}

	if (argc != 1) {
		af::array one_of_the_colors = img;
		img = af::array(one_of_the_colors.dims(0), one_of_the_colors.dims(1), 3);
		for (int i=0; i<3; i++) {
			af::array this_color;
			if (argc <= i || *argv[i] == '\0') {
				this_color = af::constant(0.0, img.dims(0), img.dims(1));
			} else {
				const char *end_of_number;
				double try_parse_as_number = std::strtod(argv[i], (char **)&end_of_number);
				if (end_of_number != NULL && *end_of_number == '\0') {
					this_color = af::constant(try_parse_as_number, img.dims(0), img.dims(1));
				} else if (i == first_non_empty_argument) {
					this_color = one_of_the_colors;
				} else {
					this_color = input_transformation(af_utils::load_cmx(argv[i]));
				}
			}
			img(af::span, af::span, i) = this_color;
		}
	}

	if (normalization == TOGETHER) {
		img = af_utils::normalize(af_utils::clampStatistically(af_utils::nonInfNaN(img), clamp_lower_sigma, clamp_upper_sigma));
	}

	if (colorspace != AF_RGB) {
		img = af::colorSpace(img, colorspace, AF_RGB);
	}

	if (output_filename.empty()) {
		char *suffix = std::strrchr(output_candidate, '.');
		output_filename = std::string(output_candidate, suffix) + ".tiff";
	}

	af::saveImage(output_filename.c_str(), img);
	return 0;
}
