#ifndef KERNELS_CUH
#define KERNELS_CUH
#include "config.hpp"
#include "common/common.cuh"

 template<int WINDOW, typename T>
__device__ inline void _smooth(T *out, const T *in, unsigned stride, unsigned ndim, unsigned ntotal) {
	for (int i = blockDim.x*blockIdx.x + threadIdx.x; i < ntotal; i += blockDim.x*gridDim.x) {
		int j = (i/stride) % ndim;

		int nbelow = WINDOW/2 + (WINDOW%2==0), nabove = WINDOW/2;
		if (j-nbelow < 0) {nbelow = j;}
		if (j+nabove >= ndim) {nabove = ndim-1-j;}

		T sum = makeFromDouble(0.0, *in);
		for (int ii=i-stride*nbelow; ii <= i+stride*nabove; ii+=stride) {
			sum = sum + in[ii];
		}
		out[i] = sum/makeFromDouble(nbelow+1+nabove, *in);
	}
}
template<int WINDOW, typename T>
__global__ void smooth(T *out, const T *in, unsigned stride, unsigned ndim, unsigned ntotal) {
	_smooth<WINDOW,T>(out, in, stride, ndim, ntotal);
}
extern "C" {__global__ void smooth_colwise(float *out, const float *in, dim3 d) {
	_smooth<WINDOW_H>(out, in, 1, d.x, d.x*d.y*d.z);
}}
extern "C" {__global__ void smooth_rowwise(float *out, const float *in, dim3 d) {
	_smooth<WINDOW_W>(out, in, d.x, d.y, d.x*d.y*d.z);
}}


struct polArray { // structures are passed by value to CUDA kernels
	float *d[NPOLARIZATION];
};
__global__ void compute_mean_mahalanobis(float *dmean_mahalanobis, const polArray dsli, const polArray dmli, dim3 d) {
	int i = blockDim.x*blockIdx.x + threadIdx.x;

	if (i < d.x*d.y) {
		float z = 0.0;
#pragma unroll
		for (int polarization=0; polarization<NPOLARIZATION; polarization++) {
			z = z + dsli.d[polarization][i]/dmli.d[polarization][i];
		}
		dmean_mahalanobis[i] = z*z;
	}
}

__global__ void compute_properties(float *dlog_rk, float *dlog_mrcs, float *dlog_xpr, const polArray dmli, const float *dmean_mahalanobis, dim3 d) {
	int i = blockDim.x*blockIdx.x + threadIdx.x;

	float cov_determinant = 1.0;
	if (i < d.x*d.y) {
#pragma unroll
		for (int polarization=0; polarization<NPOLARIZATION; polarization++) {
			cov_determinant = cov_determinant * dmli.d[polarization][i];
		}

		dlog_rk[i] = log(dmean_mahalanobis[i] / makeFromDouble(NPOLARIZATION*(NPOLARIZATION+2), *dlog_rk));
		dlog_mrcs[i] = log(cov_determinant)/ makeFromDouble(NPOLARIZATION, *dlog_mrcs);
		dlog_xpr[i] = log(dmli.d[VH][i]) - dlog_mrcs[i];
	}
}

#endif /*KERNELS_CUH*/
