#include <cassert>
#include <cstdlib>
#include <ctime>
#include <string>

#include <arrayfire.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include "backend.hpp"
#include "common/af_utils.hpp"
#include "common/monotonic.hpp"
#include "config.hpp"

char *output_prefix = NULL;

char *me;
static void die(const std::string &error = "") {
	if (!error.empty()) {
		std::cerr << "Error: " << error << std::endl;
	}
	std::cout << "Usage: "<<me<<"[-o OUTPUT_PREFIX] <WORKER> <POLARIZATION_1.cmx> ... <POLARIZATION_N.cmx>" << std::endl;
	std::exit(error.empty() ? EXIT_SUCCESS : EXIT_FAILURE);
}

enum {
	PADDING = 2*(WINDOW_W/2), // NOTE: twice the padding to envelop LOG_RK's smoothing of smoothed MLI
};

int main(int argc, char *argv[]) {
	monotonic::init();
	me = argv[0];

	while (--argc > 0 && **(++argv) == '-') {
		switch(argv[0][1]) {
			case 'h':
			case '?':
				die();

			default:
				assert(argc >= 2);
				switch(argv[0][1]) {
					case 'o': output_prefix = (*(--argc,++argv)); break;

					default:
							 die("Invalid option: \""+std::string(argv[0])+"\"");
				}
		}
	}
	if (argc < 1) die("Require "+std::to_string(1+NPOLARIZATION)+" arguments");

	af::dim4 dims = af_utils::dims_from_cmx(argv[0]);
	dim_t h = dims[0], w = dims[1];

	int fdIn[NPOLARIZATION];
	float *polarizations[NPOLARIZATION] = {NULL};
	for (int i = 0; i < NPOLARIZATION; i++) {
		fdIn[i] = open(argv[i], O_RDONLY); assert(fdIn[i] >= 0);
		polarizations[i] = (float *)mmap(NULL, dims.elements()*sizeof(**polarizations), PROT_READ, MAP_PRIVATE, fdIn[i], 0); assert(polarizations[i] != MAP_FAILED);
	}

	bool persisent_output = output_prefix != NULL;
	float *products[NPRODUCT] = {NULL};
	int fdOut[NPRODUCT];
	for (int i=0; i<NPRODUCT;i++) {
		if (persisent_output) {
			std::string fn_out = std::string(output_prefix) + "_" + product_name[i] + af_utils::cmx_suffix(f32, dims);
			fdOut[i] = open(fn_out.c_str(), O_RDWR|O_TRUNC|O_CREAT, OUTPUT_MODE); assert(fdOut[i] >= 0);
			lseek(fdOut[i], dims.elements()*sizeof(float) - 1, SEEK_SET);
			char a_byte=0;
			write(fdOut[i], &a_byte, 1);
			products[i] = (float *)mmap(NULL, dims.elements()*sizeof(**products), PROT_WRITE, MAP_SHARED, fdOut[i], 0); assert(products[i] != MAP_FAILED);
		} else {
			products[i] = (float *)malloc(h*NCOL_PER_CHUNK*(sizeof(**products))); assert(products[i] != NULL);
		}
	}

	dim_t nchunk = w / NCOL_PER_CHUNK + (w%NCOL_PER_CHUNK!=0);
	for (dim_t i=0; i<nchunk; i++) {
		double cstart = monotonic::since();
		dim_t x = i*NCOL_PER_CHUNK;
		dim_t w_before = i==0 ? 0 : std::min(dim_t(PADDING), x);
		dim_t w_after = x+NCOL_PER_CHUNK >= w ? 0 : std::min(dim_t(PADDING), w-NCOL_PER_CHUNK-x);

		float *relative_polarizations[NPOLARIZATION] = {NULL}, *relative_products[NPRODUCT] = {NULL};
		for (int p = 0; p < NPOLARIZATION; p++) {
			relative_polarizations[p] = polarizations[p] + (x-w_before)*h;
		}
		for (int p = 0; p < NPRODUCT; p++) {
			relative_products[p] = persisent_output ? products[p] + x*h : products[p];
		}

		compute_products(relative_polarizations, h, std::min(NCOL_PER_CHUNK, w-w_after-x), w_before, w_after, relative_products);
		std::printf("%lld/%lld %lf %lf\n", i, nchunk, cstart, monotonic::since());
		std::fflush(stdout);
	}

	for (int i=0; i<NPRODUCT; i++) {
		if (persisent_output) {
			//assert(msync(products[i], end_of_out*sizeof(**polarizations), MS_SYNC) == 0);
			assert(munmap(products[i], dims.elements()*sizeof(**products)) == 0);
			assert(close(fdOut[i]) == 0);
		} else {
			free(products[i]);
		}
	}

	for (int i=0; i<NPOLARIZATION; i++) {
		assert(munmap(polarizations[i], dims.elements()*sizeof(**polarizations)) == 0);
		assert(close(fdIn[i]) == 0);
	}
	return 0;
}
