#include <cassert>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <iostream>

#include "backend.hpp"
#include "config.hpp"

template<typename T>
static inline void smooth(T *out, const T *in, unsigned h, unsigned w, T *non_overlapping_tmp) {
#pragma acc kernels deviceptr(out), deviceptr(in), deviceptr(non_overlapping_tmp)
	{
#pragma acc loop independent collapse(2)
#pragma omp parallel for
		for (unsigned j=0; j<w; j++) {
			for (unsigned i=0; i<h; i++) {
				unsigned from = i - WINDOW_H/2 + (WINDOW_H%2==0), to = i+WINDOW_H/2;
				if (from > h) { from = 0; }; // NOTE: underflow
				if (to >= h) { to = h-1;};
				T sum = 0;
				for (unsigned k=from; k<=to; k++) {
					sum += in[j*h+k];
				}
				non_overlapping_tmp[j*h+i] = sum / WINDOW_H;
			}
		}
#pragma acc loop independent collapse(2)
#pragma omp parallel for
		for (unsigned j=0; j<w; j++) {
			for (unsigned i=0; i<h; i++) {
				unsigned from = j - WINDOW_W/2 + (WINDOW_W%2==0), to = j+WINDOW_W/2;
				if (from > w) { from = 0; }; // NOTE: underflow
				if (to >= w) { to = w-1;};
				T sum = 0;
				for (unsigned k=from; k<=to; k++) {
					sum += non_overlapping_tmp[k*h+i];
				}
				out[j*h+i] = sum / WINDOW_W;
			}
		}
	}
}

void compute_products(float **polarizations, size_t h, size_t w, size_t w_before, size_t w_after, float **products) {
	size_t w_in = w_before+w+w_after;
	float *mli = (float*)std::malloc(NPOLARIZATION*w_in*h*sizeof(*mli)); assert(mli != NULL);
	float *smoothed_columns = (float *)std::malloc(w_in*h*sizeof(*smoothed_columns)); assert(smoothed_columns != NULL);
	float *mli_cov_determinant = (float*)std::malloc(w_in*h*sizeof(*mli_cov_determinant)); assert(mli_cov_determinant != NULL);
	float *mean_mahalanobis = (float*)std::malloc(w_in*h*sizeof(*mean_mahalanobis)); assert(mean_mahalanobis != NULL);

#pragma acc data copyin(polarizations[:NPOLARIZATION][:w_in*h]), copyout(products[:NPRODUCT][:w*h])
#pragma acc data create(mli[:NPOLARIZATION*w_in*h]), create(smoothed_columns[:w_in*h]), create(mli_cov_determinant[:w_in*h]), create(mean_mahalanobis[:w_in*h])
	{
		for (unsigned p=0; p<NPOLARIZATION; p++) {
			const float *polarization = polarizations[p];
#pragma acc host_data use_device(mli, polarization, smoothed_columns)
			smooth(mli + p*w_in*h, polarization, h, w_in, smoothed_columns);
		}
#pragma acc kernels
		{
#pragma acc loop independent collapse(2)
#pragma omp parallel for
			for (unsigned j=0; j<w_in; j++) {
				for (unsigned i=0; i<h; i++) {
					float z = 0;
					float cov_determinant = 1;
					for (int p=0; p<NPOLARIZATION; p++) {
						z += polarizations[p][j*h + i] / mli[p*w_in*h + j*h + i];
						cov_determinant *= mli[p*w_in*h + j*h + i];
					}
					mean_mahalanobis[j*h + i] = z*z;
					mli_cov_determinant[j*h + i] = cov_determinant;
				}
			}
		}
#pragma acc host_data use_device(mean_mahalanobis, smoothed_columns)
		smooth(mean_mahalanobis, mean_mahalanobis, h, w_in, smoothed_columns);
#pragma acc kernels
		{
#pragma acc loop independent collapse(2)
#pragma omp parallel for
			for (unsigned j=0; j<w; j++) {
				for (unsigned i=0; i<h; i++) {
					unsigned j_in = w_before + j;
					float log_mrcs = std::log(mli_cov_determinant[j_in*h + i]) / NPOLARIZATION;
#pragma acc loop independent
					for (unsigned p=0; p<NPOLARIZATION; p++) {
						products[p][j*h+i] = mli[p*w_in*h + j_in*h + i];
					}
					products[LOG_RK][j*h+i] = std::log(mean_mahalanobis[j_in*h + i]) / (NPOLARIZATION*(NPOLARIZATION+2));
					products[LOG_MRCS][j*h+i] = log_mrcs;
					products[LOG_XPR][j*h+i] = std::log(mli[VH*w_in*h + j_in*h + i]) - log_mrcs;
				}
			}
		}
	}

	free(mean_mahalanobis), free(mli_cov_determinant), free(smoothed_columns), free(mli);
}
